import contextlib
import os
import subprocess
import sys
import tempfile
import time
from collections import defaultdict
from pathlib import Path
from typing import Callable
from typing import cast
from typing import Collection
from typing import Dict
from typing import FrozenSet
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple

# usually this would be a try...except ImportError
# but mypy understands only this
# see https://github.com/python/mypy/issues/1393
if sys.version_info[:2] >= (3, 8):
    # Literal is available in stdlib since Python 3.8
    from typing import Literal
else:
    from typing_extensions import Literal

import attr
import click
from attr import Factory
from wcmatch import glob as wcglob

from semgrep.config_resolver import resolve_targets
from semgrep.constants import Colors
from semgrep.error import FilesNotFoundError
from semgrep.formatter.text import width
from semgrep.ignores import FileIgnore
from semgrep.semgrep_types import FileExtension
from semgrep.semgrep_types import LANGUAGE
from semgrep.semgrep_types import Language
from semgrep.semgrep_types import LanguageDefinition
from semgrep.semgrep_types import Shebang
from semgrep.types import FilteredTargets
from semgrep.util import partition_set
from semgrep.util import sub_check_output
from semgrep.util import with_color
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)

MAX_CHARS_TO_READ_FOR_SHEBANG = 255
PATHS_ALWAYS_SKIPPED = (".git",)

ALL_EXTENSIONS: Collection[FileExtension] = {
    ext
    for definition in LANGUAGE.definition_by_id.values()
    for ext in definition.exts
    if ext != FileExtension("")
}


@contextlib.contextmanager
def converted_pipe_targets(targets: Sequence[str]) -> Iterator[Sequence[str]]:
    """
    Provides a context in which FIFOs have been copied into temp files

    This is necessary as we can not easily rewire these pipes into the called semgrep-core
    process.

    :param targets: Input target specifiers
    :return: A sequence of non-pipe specifiers (Path(t).is_file() returns true)
    """

    out_targets = []
    with tempfile.TemporaryDirectory() as temp_dir:
        for t in targets:
            if t == "-":
                with (Path(temp_dir) / "stdin").open("wb") as fd:
                    fd.write(sys.stdin.buffer.read())
                out_targets.append(fd.name)
            elif Path(t).is_fifo():
                with (Path(temp_dir) / t[1:].replace("/", "_")).open("wb") as fd:
                    with Path(t).open("rb") as td:
                        fd.write(td.read())
                out_targets.append(fd.name)
            else:
                out_targets.append(t)
        yield out_targets


@attr.s(auto_attribs=True)
class IgnoreLog:
    """Keeps track of which paths were ignored for what reason.

    Each attribute is a distinct reason why files could be ignored.

    Some reason can apply once per rule; these are mappings keyed on the rule id.
    """

    target_manager: "TargetManager"

    semgrepignored: Set[Path] = Factory(set)
    always_skipped: Set[Path] = Factory(set)
    cli_includes: Set[Path] = Factory(set)
    cli_excludes: Set[Path] = Factory(set)
    size_limit: Set[Path] = Factory(set)

    rule_includes: Dict[str, Set[Path]] = Factory(lambda: defaultdict(set))
    rule_excludes: Dict[str, Set[Path]] = Factory(lambda: defaultdict(set))

    @property
    def rule_ids_with_skipped_paths(self) -> FrozenSet[str]:
        """All rule IDs that have skipped paths.

        Note that if a rule ID defines excludes/includes,
        but they didn't skip any paths,
        that rule ID will not show up here.
        """
        return frozenset(
            {
                *(rule_id for rule_id, skips in self.rule_includes.items() if skips),
                *(rule_id for rule_id, skips in self.rule_excludes.items() if skips),
            }
        )

    def __str__(self) -> str:
        skip_fragments = []

        if self.target_manager.respect_git_ignore:
            skip_fragments.append("all .gitignored files")

        if self.cli_includes:
            skip_fragments.append(
                f"{len(self.cli_includes)} files not matching --include patterns"
            )
        if self.cli_excludes:
            skip_fragments.append(
                f"{len(self.cli_excludes)} files matching --exclude patterns"
            )
        if self.size_limit:
            skip_fragments.append(
                f"{len(self.size_limit)} files larger than {self.target_manager.max_target_bytes / 1000 / 1000} MB"
            )
        if self.semgrepignored:
            skip_fragments.append(
                f"{len(self.semgrepignored)} files matching .semgrepignore patterns"
            )

        if not skip_fragments:
            return "no files were skipped"

        message = "skipped: " + ", ".join(skip_fragments)

        message += "\nfor a detailed list of skipped files, run semgrep with the --verbose flag\n"
        return message

    def yield_verbose_lines(self) -> Iterator[Tuple[Literal[0, 1, 2], str]]:
        """Yields lines of verbose output for the skipped files.

        The returned tuple is (level, message).
        The level is a number; one of 0, 1, or 2, which sets the indentation when outputting the line.
        """
        yield 0, "Files skipped:"

        yield 1, "Always skipped by Semgrep:"
        if self.always_skipped:
            for path in self.always_skipped:
                yield 2, with_color(Colors.cyan, str(path))
        else:
            yield 2, "<none>"

        yield 1, "Skipped by .gitignore:"
        if self.target_manager.respect_git_ignore:
            yield 1, "(Disable by passing --no-git-ignore)"
            yield 2, "<all files not listed by `git ls-files` were skipped>"
        else:
            yield 1, "(Disabled with --no-git-ignore)"
            yield 2, "<none>"

        yield 1, "Skipped by .semgrepignore:"
        yield 1, "(See: https://semgrep.dev/docs/ignoring-files-folders-code/#understanding-semgrep-defaults)"
        if self.semgrepignored:
            for path in self.semgrepignored:
                yield 2, with_color(Colors.cyan, str(path))
        else:
            yield 2, "<none>"

        yield 1, "Skipped by --include patterns:"
        if self.cli_includes:
            for path in self.cli_includes:
                yield 2, with_color(Colors.cyan, str(path))
        else:
            yield 2, "<none>"

        yield 1, "Skipped by --exclude patterns:"
        if self.cli_excludes:
            for path in self.cli_excludes:
                yield 2, with_color(Colors.cyan, str(path))
        else:
            yield 2, "<none>"

        yield 1, f"Skipped by limiting to files smaller than {self.target_manager.max_target_bytes} bytes:"
        yield 1, "(Adjust with the --max-target-bytes flag)"
        if self.size_limit:
            for path in self.size_limit:
                yield 2, with_color(Colors.cyan, str(path))
        else:
            yield 2, "<none>"

        for rule_id in self.rule_ids_with_skipped_paths:
            if rule_id.startswith("fingerprints."):
                # Skip fingerprint rules, since they all have include patterns
                continue
            yield 1, f"Skipped only for {with_color(Colors.bright_blue, rule_id)} by the rule's include patterns:"
            if self.rule_includes[rule_id]:
                for path in self.rule_includes[rule_id]:
                    yield 2, with_color(Colors.cyan, str(path))
            else:
                yield 2, "<none>"

            yield 1, f"Skipped only for {with_color(Colors.bright_blue, rule_id)} by the rule's exclude patterns:"
            if self.rule_excludes[rule_id]:
                for path in self.rule_excludes[rule_id]:
                    yield 2, with_color(Colors.cyan, str(path))
            else:
                yield 2, "<none>"

    def verbose_output(self) -> str:
        formatters_by_level: Mapping[int, Callable[[str], str]] = {
            0: lambda line: "\n".join([40 * "=", line, 40 * "="]),
            1: lambda line: click.wrap_text(
                with_color(Colors.foreground, line, bold=True),
                width,
                2 * " ",
                2 * " ",
                False,
            ),
            2: lambda line: click.wrap_text(
                line,
                width,
                "   • ",
                "     ",
                False,
            ),
        }
        output = ""

        prev_level = None
        for level, line in self.yield_verbose_lines():
            if prev_level != level:
                output += "\n"
            formatter = formatters_by_level[level]
            output += formatter(line) + "\n"
            prev_level = level

        return output


@attr.s(auto_attribs=True)
class TargetManager:
    """
    Handles all file include/exclude logic for semgrep

    If respect_git_ignore is true then will only consider files that are
    tracked or (untracked but not ignored) by git

    If skip_unknown_extensions is False then targets with extensions that are
    not understood by semgrep will always be returned by get_files. Else will discard
    targets with unknown extensions

    TargetManager not to be confused with https://jobs.target.com/search-jobs/store%20manager
    """

    includes: Sequence[str]
    excludes: Sequence[str]
    max_target_bytes: int
    targets: Sequence[str] = attr.ib()
    respect_git_ignore: bool
    skip_unknown_extensions: bool
    file_ignore: Optional[FileIgnore]
    ignore_log: IgnoreLog = Factory(IgnoreLog, takes_self=True)

    _filtered_targets: Dict[Language, FilteredTargets] = attr.ib(factory=dict)

    @targets.validator
    def _check_exists(self, attribute: str, value: Sequence[str]) -> None:
        """
        Raise FilesNotFoundError if any element of targets is not a dir/regular file or
        symlink to a dir/regular file:

        i.e. does not exist, is a mount/socket etc.
        """
        targets = self.resolve_targets(self.targets)
        files, _ = partition_set(lambda p: not p.is_dir(), targets)
        _, nonexistent_files = partition_set(lambda p: p.is_file(), files)
        if nonexistent_files:
            raise FilesNotFoundError(tuple(nonexistent_files))

    @staticmethod
    def resolve_targets(targets: Sequence[str]) -> FrozenSet[Path]:
        """
        Return list of Path objects appropriately resolving relative paths
        (relative to cwd) if necessary
        """
        return frozenset(resolve_targets(targets))

    @staticmethod
    def _is_valid_file_or_dir(path: Path) -> bool:
        """Check this is a valid file or directory for semgrep scanning."""
        return os.access(str(path), os.R_OK) and not path.is_symlink()

    @staticmethod
    def _is_valid_file(path: Path) -> bool:
        """Check if file is a readable regular file.

        This eliminates files that should never be semgrep targets. Among
        others, this takes care of excluding symbolic links (because we don't
        want to scan the target twice), directories (which may be returned by
        globbing or by 'git ls-files' e.g. submodules), and files missing
        the read permission.
        """
        return TargetManager._is_valid_file_or_dir(path) and path.is_file()

    @staticmethod
    def _filter_valid_files(paths: FrozenSet[Path]) -> FrozenSet[Path]:
        """Keep only readable regular files"""
        return frozenset(path for path in paths if TargetManager._is_valid_file(path))

    @staticmethod
    def _expand_dir(
        curr_dir: Path, language: Language, respect_git_ignore: bool
    ) -> FrozenSet[Path]:
        """
        Recursively go through a directory and return list of all files with
        default file extension of language
        """

        def _parse_output(output: str, curr_dir: Path) -> FrozenSet[Path]:
            """
            Convert a newline delimited list of files to a set of path objects
            prepends curr_dir to all paths in said list

            If list is empty then returns an empty set
            """
            files: FrozenSet[Path] = frozenset()
            if output:
                files = frozenset(
                    p
                    for p in (
                        Path(curr_dir) / elem for elem in output.strip().split("\n")
                    )
                    if TargetManager._is_valid_file(p)
                )
            return files

        def _executes_with_shebang(f: Path, shebangs: Collection[Shebang]) -> bool:
            """
            Returns if a path is executable and executes with one of a set of programs
            """
            if not os.access(str(f), os.X_OK | os.R_OK):
                return False
            try:
                with f.open("r") as fd:
                    hline = fd.readline(MAX_CHARS_TO_READ_FOR_SHEBANG).rstrip()
                return any(hline.endswith(s) for s in shebangs)
            except UnicodeDecodeError:
                logger.debug(
                    f"Encountered likely binary file {f} while reading shebang; skipping this file"
                )
                return False

        def _find_files_with_extension_or_shebang(
            paths: Iterable[Path],
            definition: LanguageDefinition,
        ) -> FrozenSet[Path]:
            """
            Finds all files in a collection of paths that either:
            - end with one of a set of extensions
            - is a script that executes with one of a set of programs

            Takes ~ 100 ms to execute on a Mac PowerBook on a repo with 3000 files.
            """
            res: Set[Path] = set()
            before = time.time()
            for path in paths:
                if path.is_dir():
                    res.update(
                        p for ext in definition.exts for p in path.rglob(f"*{ext}")
                    )
                    res.update(
                        Path(root) / f
                        for root, _, files in os.walk(str(curr_dir))
                        for f in files
                        if _executes_with_shebang(Path(root) / f, definition.shebangs)
                    )
                else:
                    if any(str(path).endswith(ext) for ext in definition.exts):
                        res.add(path)
                    if _executes_with_shebang(path, definition.shebangs):
                        res.add(path)
            logger.debug(
                f"Scanned file system for matching files in {time.time() - before} s"
            )
            return frozenset(res)

        definition = LANGUAGE.definition_by_id[language]

        if respect_git_ignore:
            try:
                # git ls-files is significantly faster than os.walk when performed on a git project,
                # so identify the git files first, then filter those

                # Tracked files
                tracked_output = sub_check_output(
                    ["git", "ls-files"],
                    cwd=curr_dir.resolve(),
                    encoding="utf-8",
                    stderr=subprocess.DEVNULL,
                )

                # Untracked but not ignored files
                untracked_output = sub_check_output(
                    [
                        "git",
                        "ls-files",
                        "--other",
                        "--exclude-standard",
                    ],
                    cwd=curr_dir.resolve(),
                    encoding="utf-8",
                    stderr=subprocess.DEVNULL,
                )

                deleted_output = sub_check_output(
                    ["git", "ls-files", "--deleted"],
                    cwd=curr_dir.resolve(),
                    encoding="utf-8",
                    stderr=subprocess.DEVNULL,
                )
                tracked = _parse_output(tracked_output, curr_dir)
                untracked_unignored = _parse_output(untracked_output, curr_dir)
                deleted = _parse_output(deleted_output, curr_dir)
                paths = tracked.union(untracked_unignored).difference(deleted)
                results = _find_files_with_extension_or_shebang(paths, definition)

            except (subprocess.CalledProcessError, FileNotFoundError):
                logger.verbose(
                    f"Unable to ignore files ignored by git ({curr_dir} is not a git directory or git is not installed). Running on all files instead..."
                )
                # Not a git directory or git not installed. Fallback to using rglob
                results = _find_files_with_extension_or_shebang([curr_dir], definition)
        else:
            results = _find_files_with_extension_or_shebang([curr_dir], definition)

        return TargetManager._filter_valid_files(results)

    @staticmethod
    def expand_targets(
        targets: Collection[Path], lang: Language, respect_git_ignore: bool
    ) -> FrozenSet[Path]:
        """
        Explore all directories. Remove duplicates
        """
        expanded: Set[Path] = set()
        for target in targets:
            if not TargetManager._is_valid_file_or_dir(target):
                continue

            if target.is_dir():
                expanded.update(
                    TargetManager._expand_dir(target, lang, respect_git_ignore)
                )
            else:
                expanded.add(target)

        return frozenset(expanded)

    @staticmethod
    def preprocess_path_patterns(patterns: Sequence[str]) -> List[str]:
        """Convert semgrep's path include/exclude patterns to wcmatch's glob patterns.

        In semgrep, pattern "foo/bar" should match paths "x/foo/bar", "foo/bar/x", and
        "x/foo/bar/x". It implicitly matches zero or more directories at the beginning and the end
        of the pattern. In contrast, we have to explicitly specify the globstar (**) patterns in
        wcmatch. This function will converts a pattern "foo/bar" into "**/foo/bar" and
        "**/foo/bar/**". We need the pattern without the trailing "/**" because "foo/bar.py/**"
        won't match "foo/bar.py".
        """
        result = []
        for pattern in patterns:
            result.append("**/" + pattern)
            result.append("**/" + pattern + "/**")
        return result

    @staticmethod
    def filter_includes(
        includes: Sequence[str], *, candidates: FrozenSet[Path]
    ) -> FilteredTargets:
        """
        Returns all elements in candidates that match any includes pattern

        If includes is empty, returns candidates unchanged
        """
        if not includes:
            return FilteredTargets(candidates)

        includes = TargetManager.preprocess_path_patterns(includes)
        # Need cast b/c types-wcmatch doesn't use generics properly :(
        kept = frozenset(
            cast(
                Iterable[Path],
                wcglob.globfilter(
                    candidates, includes, flags=wcglob.GLOBSTAR | wcglob.DOTGLOB
                ),
            )
        )
        return FilteredTargets(kept, frozenset(candidates - kept))

    @staticmethod
    def filter_excludes(
        excludes: Sequence[str], *, candidates: FrozenSet[Path]
    ) -> FilteredTargets:
        """
        Returns all elements in candidates that do not match any excludes pattern

        If excludes is empty, returns candidates unchanged
        """
        if not excludes:
            return FilteredTargets(candidates)

        excludes = TargetManager.preprocess_path_patterns(excludes)

        # Need cast b/c types-wcmatch doesn't use generics properly :(
        removed = frozenset(
            cast(
                Iterable[Path],
                wcglob.globfilter(
                    candidates, excludes, flags=wcglob.GLOBSTAR | wcglob.DOTGLOB
                ),
            )
        )
        return FilteredTargets(frozenset(candidates - removed), removed)

    @staticmethod
    def filter_by_size(
        max_target_bytes: int, *, candidates: FrozenSet[Path]
    ) -> FilteredTargets:
        """
        Return all the files whose size doesn't exceed the limit.

        If max_target_bytes is zero or negative, all paths are returned.
        If some paths are invalid, they may or may not be included in the
        result.
        """
        if max_target_bytes <= 0:
            return FilteredTargets(candidates)

        kept, removed = partition_set(
            lambda path: TargetManager._is_valid_file(path)
            and os.path.getsize(path) <= max_target_bytes,
            candidates,
        )

        return FilteredTargets(kept, removed)

    def filtered_files(self, lang: Language) -> FilteredTargets:
        """
        Return all files that are decendants of any directory in TARGET that have
        an extension matching LANG that match any pattern in INCLUDES and do not
        match any pattern in EXCLUDES. Any file in TARGET bypasses excludes and includes.
        If a file in TARGET has a known extension that is not for langugage LANG then
        it is also filtered out

        Note also filters out any directory and decendants of `.git`
        """
        if lang in self._filtered_targets:
            return self._filtered_targets[lang]

        # Non dir/non file should not exist cause of init time validation
        # See _check_exists()
        discovered_paths = self.resolve_targets(self.targets)
        files, directories = partition_set(lambda p: not p.is_dir(), discovered_paths)
        explicit_files, _ = partition_set(lambda p: p.is_file(), files)

        discovered_paths = self.expand_targets(
            directories, lang, self.respect_git_ignore
        )

        filtered_targets = self.filter_includes(
            self.includes,
            candidates=discovered_paths,
        )
        self.ignore_log.cli_includes.update(filtered_targets.removed)

        filtered_targets = self.filter_excludes(
            self.excludes,
            candidates=filtered_targets.kept,
        )
        self.ignore_log.cli_excludes.update(filtered_targets.removed)

        filtered_targets = self.filter_excludes(
            PATHS_ALWAYS_SKIPPED,
            candidates=filtered_targets.kept,
        )
        self.ignore_log.always_skipped.update(filtered_targets.removed)

        filtered_targets = self.filter_by_size(
            self.max_target_bytes,
            candidates=filtered_targets.kept,
        )
        self.ignore_log.size_limit.update(filtered_targets.removed)

        # Remove explicit_files with known extensions.
        explicit_files_with_lang_extension = frozenset(
            f
            for f in explicit_files
            if (any(f.match(f"*{ext}") for ext in LANGUAGE.definition_by_id[lang].exts))
        )
        targets = filtered_targets.kept.union(explicit_files_with_lang_extension)

        if not self.skip_unknown_extensions:
            explicit_files_with_unknown_extensions = frozenset(
                f
                for f in explicit_files
                if not any(f.match(f"*{ext}") for ext in ALL_EXTENSIONS)
            )
            targets = targets.union(explicit_files_with_unknown_extensions)

        filtered_targets = (
            self.file_ignore.filter_paths(candidates=targets)
            if self.file_ignore
            else FilteredTargets(targets)
        )
        self.ignore_log.semgrepignored.update(filtered_targets.removed)

        self._filtered_targets[lang] = FilteredTargets(
            filtered_targets.kept,
            discovered_paths - filtered_targets.kept,
        )
        return self._filtered_targets[lang]

    def get_files(
        self,
        lang: Language,
        includes: Sequence[str],
        excludes: Sequence[str],
        rule_id: str,
    ) -> FrozenSet[Path]:
        """
        Returns list of files that should be analyzed for a LANG

        Given this object's TARGET, self.INCLUDE, and self.EXCLUDE will return list
        of all descendant files of directories in TARGET that end in extension
        typical for LANG. If self.INCLUDES is non empty then all files will have an ancestor
        that matches a pattern in self.INCLUDES. Will not include any file that has
        an ancestor that matches a pattern in self.EXCLUDES. Any explicitly named files
        in TARGET will bypass this global INCLUDE/EXCLUDE filter. The local INCLUDE/EXCLUDE
        filter is then applied.
        """
        filtered_targets = self.filtered_files(lang)

        filtered_targets = self.filter_includes(
            includes, candidates=filtered_targets.kept
        )
        self.ignore_log.rule_includes[rule_id].update(filtered_targets.removed)

        filtered_targets = self.filter_excludes(
            excludes, candidates=filtered_targets.kept
        )
        self.ignore_log.rule_excludes[rule_id].update(filtered_targets.removed)

        filtered_targets = self.filter_by_size(
            self.max_target_bytes, candidates=filtered_targets.kept
        )
        self.ignore_log.size_limit.update(filtered_targets.removed)

        return filtered_targets.kept



"""
Handles ignoring Semgrep findings via inline code comments

Currently supports ignoring a finding on a single line by adding a
`# nosemgrep:ruleid` comment (or `// nosemgrep:ruleid`).

To use, create a RuleMatchMap, then pass it to process_ignores().

Coupling: semgrep-action uses a regexp to strip nosemgrep comments so as
to normalize the code and not be sensitive to the addition or removal
of nosemgrep comments.
See https://github.com/returntocorp/semgrep-action/blob/develop/src/semgrep_agent/findings.py
and check that it's compatible with any change we're making here.
"""
from re import sub
from typing import List
from typing import Sequence
from typing import Tuple

import attr

from semgrep.constants import COMMA_SEPARATED_LIST_RE
from semgrep.constants import NOSEM_INLINE_RE
from semgrep.constants import NOSEM_PREVIOUS_LINE_RE
from semgrep.error import Level
from semgrep.error import SemgrepError
from semgrep.rule_match import RuleMatch
from semgrep.rule_match_map import RuleMatchMap
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)


def process_ignores(
    rule_matches_by_rule: RuleMatchMap,
    keep_ignored: bool,
    *,
    strict: bool,
) -> Tuple[RuleMatchMap, Sequence[SemgrepError], int]:
    """
    Converts a mapping of findings to a mapping of findings that
    will be shown to the caller.

    :param rule_matches_by_rule: The input findings (typically from a Semgrep call)
    :param keep_ignored: if true will keep nosem findings in returned object, otherwise removes them
    :param strict: The value of the --strict flag (affects error return)
    :return:
    - RuleMatchMap: dict from rule to list of findings. Findings have is_ignored
        set to true if there was matching nosem comment found for it.
        If keep_ignored set to true, will keep all findings that have is_ignored: True
        otherwise removes them in the return object
    - list of semgrep errors when dealing with nosem:
        i.e. a nosem without associated finding or nosem id not matching finding
    - number of findings with is_ignored set to true
    """
    filtered = {}
    nosem_errors: List[SemgrepError] = []
    for rule, matches in rule_matches_by_rule.items():
        evolved_matches = []
        for match in matches:
            ignored, returned_errors = _rule_match_nosem(match, strict)
            evolved_matches.append(attr.evolve(match, is_ignored=ignored))
            nosem_errors.extend(returned_errors)
        filtered[rule] = evolved_matches

    if not keep_ignored:
        filtered = {
            rule: [m for m in matches if not m._is_ignored]
            for rule, matches in filtered.items()
        }

    num_findings_nosem = sum(
        1 for rule, matches in filtered.items() for m in matches if m._is_ignored
    )

    return filtered, nosem_errors, num_findings_nosem


def _rule_match_nosem(
    rule_match: RuleMatch, strict: bool
) -> Tuple[bool, Sequence[SemgrepError]]:
    if not rule_match.lines:
        return False, []

    ids: List[str] = []

    # Only consider the first line of a match. This will keep consistent
    # behavior on where we expect a 'nosem' comment to exist. If we allow these
    # comments on any line of a match it will get confusing as to what finding
    # the 'nosem' is referring to.
    lines_re_match = NOSEM_INLINE_RE.search(rule_match.lines[0])
    if lines_re_match:
        lines_ids_str = lines_re_match.groupdict()["ids"]
        if lines_ids_str:
            ids = ids + COMMA_SEPARATED_LIST_RE.split(lines_ids_str)

    # Same thing, but inspect the line just before the match.
    # This involves a different regexp. Also captures rule names as 'ids'.
    prev_line_re_match = NOSEM_PREVIOUS_LINE_RE.search(rule_match.previous_line)
    if prev_line_re_match:
        prev_line_ids_str = prev_line_re_match.groupdict()["ids"]
        if prev_line_ids_str:
            ids = ids + COMMA_SEPARATED_LIST_RE.split(prev_line_ids_str)

    if lines_re_match is None and prev_line_re_match is None:
        return False, []

    if not ids:
        logger.verbose(
            f"found 'nosem' comment, skipping rule '{rule_match.id}' on line {rule_match.start.line}"
        )
        return True, []

    # Strip quotes to allow for use of nosem as an HTML attribute inside tags.
    # HTML comments inside tags are not allowed by the spec.
    pattern_ids = {
        pattern_id.strip().strip("\"'") for pattern_id in ids if pattern_id.strip()
    }

    # Filter out ids that are not alphanum+dashes+underscores+periods.
    # This removes trailing symbols from comments, such as HTML comments `-->`
    # or C-like multiline comments `*/`.
    pattern_ids = set(filter(lambda x: not sub(r"[\w\-\.]+", "", x), pattern_ids))

    errors = []
    result = False
    for pattern_id in pattern_ids:
        if rule_match.id == pattern_id:
            logger.verbose(
                f"found 'nosem' comment with id '{pattern_id}', skipping rule '{rule_match.id}' on line {rule_match.start.line}"
            )
            result = result or True
        else:
            message = f"found 'nosem' comment with id '{pattern_id}', but no corresponding rule trying '{rule_match.id}'"
            if strict:
                errors.append(SemgrepError(message, level=Level.WARN))
            else:
                logger.verbose(message)

    return result, errors



import base64
import collections
import json
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Generator
from typing import List

from semgrep.error import SemgrepError
from semgrep.verbose_logging import getLogger


logger = getLogger(__name__)

from dependencyparser.models import LockfileDependency
from dependencyparser.models import PackageManagers


def parse_lockfile_str(
    lockfile_text: str, filepath_for_reference: Path
) -> Generator[LockfileDependency, None, None]:
    # coupling with the github action, which decides to send files with these names back to us
    if filepath_for_reference.name.lower() == "pipfile.lock":
        yield from parse_Pipfile_str(lockfile_text)
    elif filepath_for_reference.name.lower() == "yarn.lock":
        yield from parse_Yarnlock_str(lockfile_text)
    elif filepath_for_reference.name.lower() == "package-lock.json":
        yield from parse_NPM_package_lock_str(lockfile_text)
    else:
        raise SemgrepError(
            f"don't know how to parse this filename: {filepath_for_reference}"
        )


def extract_npm_lockfile_hash(s: str) -> Dict[str, List[str]]:
    """
    Go from:
        sha512-aePbxDmcYW++PaqBsJ+HYUFwCdv4LVvdnhBy78E57PIor8/OVvhMrADFFEDh8DHDFRv/O9i3lPhsENjO7QX0+A==
    To:
        sha512,
    """
    algorithm = s.split("-")[0]
    rest = s[len(algorithm) + 1 :]
    decode_base_64 = base64.b64decode(rest)
    return {algorithm: [base64.b16encode(decode_base_64).decode("ascii").lower()]}


def parse_Yarnlock_str(lockfile_text: str) -> Generator[LockfileDependency, None, None]:
    def extract_yarn_name(line: str) -> str:
        """
        "@babel/code-frame@^7.0.0", "@babel/code-frame@^7.10.4", "@babel/code-frame@^7.8.3":
        should parse as @babel/code-frame

        "@babel/types@^7.0.0", "@babel/types@^7.10.4", "@babel/types@^7.10.5", "@babel/types@^7.11.0", "@babel/types@^7.3.0", "@babel/types@^7.3.3", "@babel/types@^7.4.0", "@babel/types@^7.4.4", "@babel/types@^7.7.0", "@babel/types@^7.9.0":
        should parse as @babel/types

        lodash@4.17.18:
        should parse as lodash
        """
        # print(line)
        if '"' not in line:
            return line.split("@")[0]
        else:
            first_quoted = "".join(line.split('"')[1:2])
            parsed_name = "@".join(first_quoted.split("@")[:-1])
            print(line, first_quoted, parsed_name)
            return parsed_name

    def remove_trailing_octothorpe(s: str) -> str:
        return "#".join(s.split("#")[:-1])

    package_name, version, resolved, integrity = None, None, None, None
    for line in lockfile_text.split("\n") + [""]:
        line = line.strip()
        # print(line)
        # print('>>>', package_name, version, resolved)
        if line.startswith("#"):
            continue
        if package_name is None and len(line) != 0:
            package_name = extract_yarn_name(line)
            continue

        if line.startswith("version") and not version:
            version = line.split("version")[1].replace('"', "").strip()
            continue

        elif line.startswith("resolved") and not resolved:
            resolved = remove_trailing_octothorpe(
                line.split("resolved")[1].replace('"', "").strip()
            )
            continue

        elif line.startswith("integrity") and not integrity:
            integrity = line.split("integrity")[1].replace('"', "").strip()
            continue

        if len(line) == 0 and package_name and version and resolved and integrity:
            yield LockfileDependency(
                package_name,
                version,
                PackageManagers.NPM,
                allowed_hashes=extract_npm_lockfile_hash(integrity),
                resolved_url=[resolved],
            )
            package_name, version, resolved, integrity = None, None, None, None


def parse_NPM_package_lock_str(
    lockfile_text: str,
) -> Generator[LockfileDependency, None, None]:
    as_json = json.loads(lockfile_text)
    for dep in as_json["dependencies"]:
        dep_blob = as_json["dependencies"][dep]
        version = dep_blob.get("version", None)
        if not version:
            logger.info(f"no version for dependency: {dep}")
        else:
            resolved_url = dep_blob.get("resolved")
            integrity = dep_blob["integrity"]
            yield LockfileDependency(
                dep,
                version,
                PackageManagers.NPM,
                allowed_hashes=extract_npm_lockfile_hash(integrity),
                resolved_url=[resolved_url] if resolved_url else None,
            )


def parse_Pipfile_str(lockfile_text: str) -> Generator[LockfileDependency, None, None]:
    def extract_pipfile_hashes(
        hashes: List[str],
    ) -> Dict[str, List[str]]:
        output = collections.defaultdict(list)
        for h in hashes:
            algorithm = h.split(":")[0]
            rest = h[len(algorithm) + 1 :]  # pipfile is already in base16
            output[algorithm].append(rest.lower())
        return output

    def parse_dependency_blob(
        root_blob: Dict[str, Any]
    ) -> Generator[LockfileDependency, None, None]:
        for dep in root_blob:
            dep_blob = root_blob[dep]
            version = dep_blob.get("version", None)
            if not version:
                logger.info(f"no version for dependency: {dep}")
            else:
                version = version.replace("==", "")
                yield LockfileDependency(
                    dep,
                    version,
                    PackageManagers.PYPI,
                    resolved_url=None,
                    allowed_hashes=extract_pipfile_hashes(dep_blob["hashes"]),
                )

    as_json = json.loads(lockfile_text)
    yield from parse_dependency_blob(as_json["default"])
    develop_deps = as_json.get("develop")
    if develop_deps is not None:
        yield from parse_dependency_blob(develop_deps)



from unittest.mock import patch

from click.testing import CliRunner

from semgrep.cli import cli
from semgrep.commands.login import Authentication
from semgrep.constants import SEMGREP_SETTING_ENVVAR_NAME


def test_login(tmp_path):
    runner = CliRunner(env={SEMGREP_SETTING_ENVVAR_NAME: str(tmp_path)})

    # Patch Token Validation:
    with patch.object(Authentication, "is_valid_token", return_value=True):
        # Logout
        result = runner.invoke(
            cli,
            ["logout"],
        )
        assert result.exit_code == 0
        assert result.output == "logged out\n"

        # Login
        result = runner.invoke(
            cli,
            ["login"],
            input="key123",
        )
        print(result.output)
        assert result.exit_code == 0
        assert "Valid API Token saved in" in result.output

        # Login should fail on second call
        result = runner.invoke(
            cli,
            ["login"],
        )
        assert result.exit_code == 1
        assert "API token already exists in" in result.output

        # Clear login
        result = runner.invoke(
            cli,
            ["logout"],
        )
        assert result.exit_code == 0
        assert result.output == "logged out\n"

        # Logout twice should work
        result = runner.invoke(
            cli,
            ["logout"],
        )
        assert result.exit_code == 0
        assert result.output == "logged out\n"

        # Check registry config works while logged in
        # Login
        result = runner.invoke(
            cli,
            ["login"],
            input="key123",
        )
        print(result.output)
        assert result.exit_code == 0
        assert "Valid API Token saved in" in result.output

        # Run p/ci
        result = runner.invoke(
            cli,
            ["--config", "r/python.lang.correctness.useless-eqeq.useless-eqeq"],
        )
        assert result.exit_code == 0

        # Run policy with bad token -> no associated deployment_id
        result = runner.invoke(cli, ["--config", "policy"])
        assert result.exit_code == 1
        assert "Invalid API Key" in result.output

        with patch.object(Authentication, "get_deployment_id", return_value=1):
            # Run policy without SEMGREP_REPO_NAME
            result = runner.invoke(
                cli,
                ["--config", "policy"],
            )
            assert result.exit_code == 1
            assert "Need to set env var SEMGREP_REPO_NAME" in result.output

            # Run policy. Check that request is made to correct deployment + org/repo
            result = runner.invoke(
                cli, ["--config", "policy"], env={"SEMGREP_REPO_NAME": "org/repo"}
            )
            assert result.exit_code == 1
            assert (
                "https://semgrep.dev/api/agent/deployment/1/repos/org/repo/rules.yaml"
                in result.output
            )



import json
from subprocess import CalledProcessError

import pytest


@pytest.mark.parametrize(
    "rule,target",
    [
        ("rules/invalid-rules/invalid-metavariable-regex.yaml", "basic/stupid.py"),
        ("rules/invalid-rules/invalid-pattern-child.yaml", "basic/stupid.py"),
        ("rules/invalid-rules/invalid-missing-top-item.yaml", "basic/stupid.py"),
    ],
)
def test_validation_of_invalid_rules(run_semgrep_in_tmp, snapshot, rule, target):
    with pytest.raises(CalledProcessError) as exc_info:
        run_semgrep_in_tmp(rule, target_name=target)

    semgrep_json_output = json.loads(exc_info.value.stdout)

    snapshot.assert_match(
        json.dumps(semgrep_json_output, indent=2, sort_keys=True), "results.json"
    )



#!/usr/bin/env python3
import sys

from semgrep import __VERSION__
from semgrep.cli import cli
from semgrep.error import OK_EXIT_CODE
from semgrep.error import SemgrepError
from semgrep.metric_manager import metric_manager
from semgrep.verbose_logging import getLogger


def main() -> int:
    # When running semgrep as a command line tool
    # silence root level logger otherwise logs higher
    # than warning are handled twice
    logger = getLogger("semgrep")
    logger.propagate = False
    metric_manager.set_version(__VERSION__)
    try:
        cli()
    # Catch custom exceptions, output the right message and exit.
    # Note: this doesn't catch all Exceptions and lets them bubble up.
    except SemgrepError as e:
        metric_manager.set_return_code(e.code)
        return e.code
    else:
        metric_manager.set_return_code(OK_EXIT_CODE)
        return OK_EXIT_CODE
    finally:
        metric_manager.send()


if __name__ == "__main__":
    sys.exit(main())



"""
This file encapsulates classes necessary in parsing semgrep-core json output into a typed object

The objects in this class should expose functionality that returns objects that the rest of the codebase
interacts with (e.g. the rest of the codebase should be interacting with RuleMatch objects instead of CoreMatch
and SemgrepCoreError instead of CoreError objects).

The precise type of the response from semgrep-core is specified in
Semgrep_core_response.atd, currently at:
https://github.com/returntocorp/semgrep/blob/develop/semgrep-core/src/core-response/Semgrep_core_response.atd
"""
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import NewType
from typing import Optional
from typing import Tuple

import attr

from semgrep.error import LegacySpan
from semgrep.error import Level
from semgrep.error import SemgrepCoreError
from semgrep.rule import Rule
from semgrep.rule_match import CoreLocation
from semgrep.rule_match import RuleMatch
from semgrep.types import JsonObject
from semgrep.types import RuleId
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)

CoreErrorMessage = NewType("CoreErrorMessage", str)
CoreErrorType = NewType("CoreErrorType", str)
SkipReason = NewType("SkipReason", str)
SkipDetails = NewType("SkipDetails", str)

CoreRulesParseTime = NewType("CoreRulesParseTime", float)


@attr.s(auto_attribs=True, frozen=True)
class MetavarValue:
    start: CoreLocation
    end: CoreLocation

    @classmethod
    def parse(cls, raw_json: JsonObject) -> "MetavarValue":
        start = CoreLocation.parse(raw_json["start"])
        end = CoreLocation.parse(raw_json["end"])
        return cls(start, end)


@attr.s(auto_attribs=True, frozen=True)
class CoreMetavars:
    metavars: Dict[str, MetavarValue]

    @classmethod
    def parse(cls, raw_json: JsonObject) -> "CoreMetavars":
        metavars: Dict[str, MetavarValue] = {}
        for key, value in raw_json.items():
            metavars[key] = MetavarValue.parse(value)
        return cls(metavars)

    def get(self, key: str) -> MetavarValue:
        return self.metavars[key]

    def keys(self) -> List[str]:
        return list(self.metavars.keys())


@attr.s(auto_attribs=True, frozen=True)
class CoreMatch:
    """
    Encapsulates finding returned by semgrep-core
    """

    rule: Rule
    path: Path
    start: CoreLocation
    end: CoreLocation
    extra: Dict[str, Any]
    metavars: CoreMetavars

    @classmethod
    def parse(cls, rule_table: Dict[RuleId, Rule], raw_json: JsonObject) -> "CoreMatch":
        rule_id = rule_table[RuleId(raw_json["rule_id"])]
        location = raw_json["location"]
        path_str = location["path"]
        assert isinstance(path_str, str)
        path = Path(path_str)
        start = CoreLocation.parse(location["start"])
        end = CoreLocation.parse(location["end"])
        extra = raw_json.get("extra", {})
        metavars = CoreMetavars.parse(extra.get("metavars"))
        return cls(rule_id, path, start, end, extra, metavars)


@attr.s(auto_attribs=True, frozen=True)
class CoreError:
    """
    Encapsulates error object returned by semgrep-core
    and handles conversion into SemgrepCoreError class that rest of codebase understands.
    """

    error_type: CoreErrorType
    rule_id: Optional[RuleId]
    path: Path
    start: CoreLocation
    end: CoreLocation
    message: CoreErrorMessage
    level: Level
    spans: Optional[Tuple[LegacySpan, ...]]
    details: Optional[str]

    @classmethod
    def parse(cls, raw_json: JsonObject) -> "CoreError":
        error_type = CoreErrorType(raw_json["error_type"])
        raw_rule_id = raw_json.get("rule_id")
        rule_id = RuleId(raw_rule_id) if raw_rule_id else None
        location = raw_json["location"]
        path = Path(location["path"])
        start = CoreLocation.parse(location["start"])
        end = CoreLocation.parse(location["end"])
        _extra = raw_json.get("extra", {})
        message = CoreErrorMessage(raw_json.get("message", "<no error message>"))
        level_str = raw_json["severity"]
        if level_str.upper() == "WARNING":
            level_str = "WARN"
        level = Level[level_str.upper()]

        details = raw_json.get("details")

        # TODO legacy support for live editor pattern parse highlighting
        spans = None
        if "yaml_path" in raw_json:
            yaml_path = tuple(raw_json["yaml_path"])
            yaml_path = yaml_path[::-1]
            spans = tuple([LegacySpan(start, end, yaml_path)])  # type: ignore

        return cls(
            error_type, rule_id, path, start, end, message, level, spans, details
        )

    def is_timeout(self) -> bool:
        """
        Return if this error is a match timeout
        """
        return self.error_type == CoreErrorType("Timeout")

    def to_semgrep_error(self) -> SemgrepCoreError:
        reported_rule_id = self.rule_id

        # TODO benchmarking code relies on error code value right now
        # See https://semgrep.dev/docs/cli-usage/ for meaning of codes
        if self.error_type == CoreErrorType(
            "Syntax error"
        ) or self.error_type == CoreErrorType("Lexical error"):
            code = 3
            reported_rule_id = None  # Rule id not important for parse errors
        else:
            code = 2

        return SemgrepCoreError(
            code,
            self.level,
            self.error_type,
            reported_rule_id,
            self.path,
            self.start,
            self.end,
            self.message,
            self.spans,
            self.details,
        )


@attr.s(auto_attribs=True, frozen=True)
class CoreSkipped:
    rule_id: Optional[RuleId]
    path: Path
    reason: SkipReason
    details: SkipDetails

    @classmethod
    def parse(cls, raw_json: JsonObject) -> "CoreSkipped":
        raw_rule_id = raw_json.get("rule_id")
        rule_id = RuleId(raw_rule_id) if raw_rule_id else None
        path = Path(raw_json["path"])
        reason = SkipReason(raw_json["reason"])
        details = SkipDetails(raw_json["details"])
        if raw_rule_id:
            rule_info = f"rule {raw_rule_id}"
        else:
            rule_info = "all rules"
        logger.verbose(f"skipped '{path}' [{rule_info}]: {reason}: {details}")
        return cls(rule_id, path, reason, details)


@attr.s(auto_attribs=True, frozen=True)
class CoreRuleTiming:  # For a given target
    rule: Rule
    parse_time: float
    match_time: float

    @classmethod
    def parse(
        cls, rule_table: Dict[RuleId, Rule], raw_json: JsonObject
    ) -> "CoreRuleTiming":
        rule = rule_table[RuleId(raw_json["rule_id"])]
        parse_time = raw_json["parse_time"]
        match_time = raw_json["match_time"]
        return cls(rule, parse_time, match_time)


@attr.s(auto_attribs=True, frozen=True)
class CoreTargetTiming:
    target: Path
    per_rule_timings: List[CoreRuleTiming]
    run_time: float

    @classmethod
    def parse(
        cls, rule_table: Dict[RuleId, Rule], raw_json: JsonObject
    ) -> "CoreTargetTiming":
        target = Path(raw_json["path"])
        per_rule_timings = [
            CoreRuleTiming.parse(rule_table, timing)
            for timing in raw_json["rule_times"]
        ]
        run_time = raw_json["run_time"]
        return cls(target, per_rule_timings, run_time)


@attr.s(auto_attribs=True, frozen=True)
class CoreTiming:
    rules: List[Rule]
    target_timings: List[CoreTargetTiming]
    rules_parse_time: CoreRulesParseTime

    @classmethod
    def parse(
        cls, rule_table: Dict[RuleId, Rule], raw_json: JsonObject
    ) -> "CoreTiming":
        if not raw_json:
            return cls([], [], CoreRulesParseTime(0.0))

        rules = [rule_table[RuleId(rule)] for rule in raw_json.get("rules", [])]
        target_timings = [
            CoreTargetTiming.parse(rule_table, target)
            for target in raw_json.get("targets", [])
        ]
        rules_parse_time = raw_json.get("rules_parse_time", 0.0)

        return cls(rules, target_timings, rules_parse_time)


@attr.s(auto_attribs=True)
class CoreOutput:
    """
    Parses output of semgrep-core
    """

    matches: List[CoreMatch]
    errors: List[CoreError]
    skipped: List[CoreSkipped]
    timing: CoreTiming

    @classmethod
    def parse(cls, rules: List[Rule], raw_json: JsonObject) -> "CoreOutput":
        rule_table = {RuleId(rule.id): rule for rule in rules}

        parsed_errors = []
        errors = raw_json["errors"]
        for error in errors:
            parsed_errors.append(CoreError.parse(error))

        parsed_matches = []
        matches = raw_json["matches"]
        for match in matches:
            parsed_matches.append(CoreMatch.parse(rule_table, match))

        parsed_skipped = []
        skipped = raw_json["skipped"]
        for skip in skipped:
            parsed_skipped.append(CoreSkipped.parse(skip))

        timings = raw_json.get("time", {})
        parsed_timings = CoreTiming.parse(rule_table, timings)

        return cls(parsed_matches, parsed_errors, parsed_skipped, parsed_timings)

    def rule_matches(self, rules: List[Rule]) -> Dict[Rule, List[RuleMatch]]:
        """
        Convert core_match objects into RuleMatch objects that the rest of the codebase
        interacts with.

        For now assumes that all matches encapsulated by this object are from the same rulee
        """
        # This will remove matches that have the same range but different
        # metavariable bindings, choosing the last one in the list. We want the
        # last because if there multiple possible bindings, they will be returned
        # by semgrep-core from largest range to smallest. For an example, see
        # tests/e2e/test_message_interpolation.py::test_message_interpolation;
        # specifically, the multi-pattern-inside test
        #
        # Another option is to not dedup, since Semgrep.ml now does its own deduping
        # otherwise, and surface both matches
        def dedup(outputs: List[RuleMatch]) -> List[RuleMatch]:
            return list({uniq_id(r): r for r in reversed(outputs)}.values())[::-1]

        def uniq_id(
            r: RuleMatch,
        ) -> Tuple[str, Path, int, int, str]:
            start = r.start
            end = r.end
            return (
                r.id,
                r.path,
                start.offset,
                end.offset,
                r.message,
            )

        def interpolate(text: str, metavariables: Dict[str, str]) -> str:
            """Interpolates a string with the metavariables contained in it, returning a new string"""

            # Sort by metavariable length to avoid name collisions (eg. $X2 must be handled before $X)
            for metavariable in sorted(metavariables.keys(), key=len, reverse=True):
                text = text.replace(metavariable, metavariables[metavariable])

            return text

        def read_metavariables(match: CoreMatch) -> Dict[str, str]:
            result = {}

            # open path and ignore non-utf8 bytes. https://stackoverflow.com/a/56441652
            with open(match.path, errors="replace") as fd:
                for metavariable in match.metavars.keys():
                    metavariable_data = match.metavars.get(metavariable)
                    # Offsets are start inclusive and end exclusive
                    start_offset = metavariable_data.start.offset
                    end_offset = metavariable_data.end.offset
                    length = end_offset - start_offset

                    fd.seek(start_offset)
                    result[metavariable] = fd.read(length)

            return result

        def convert_to_rule_match(match: CoreMatch) -> RuleMatch:
            rule = match.rule
            metavariables = read_metavariables(match)
            message = interpolate(rule.message, metavariables)
            fix = interpolate(rule.fix, metavariables) if rule.fix else None

            rule_match = RuleMatch(
                rule._id,
                message=message,
                metadata=rule.metadata,
                severity=rule.severity,
                fix=fix,
                fix_regex=rule.fix_regex,
                path=match.path,
                start=match.start,
                end=match.end,
                extra=match.extra,
            )
            return rule_match

        def order_rule_matches(matches: List[RuleMatch]) -> List[RuleMatch]:
            sorted_matches = sorted(
                matches,
                key=lambda rule_match: [rule_match.path, rule_match.start.offset],
            )
            return dedup(sorted_matches)

        findings: Dict[Rule, List[RuleMatch]] = {rule: [] for rule in rules}
        for match in self.matches:
            rule_match = convert_to_rule_match(match)
            findings[match.rule].append(rule_match)

        # Sort results so as to guarantee the same results across different
        # runs. Results may arrive in a different order due to parallelism
        # (-j option).
        findings = {
            rule: order_rule_matches(matches) for rule, matches in findings.items()
        }
        return findings



import os
import sys
from typing import Optional

import click

from semgrep.constants import SEMGREP_URL
from semgrep.constants import SEMGREP_USER_AGENT
from semgrep.settings import SETTINGS
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)


@click.command()
def login() -> None:
    """
    Prompts for API key to semgrep.dev and saves it to global settings file
    """
    saved_login_token = Authentication.read_token()
    if saved_login_token:
        click.echo(
            f"API token already exists in {SETTINGS.get_path_to_settings()}. To login with a different token logout with `semgrep logout`"
        )
        sys.exit(1)

    login_token = click.prompt("Enter semgrep.dev API token", hide_input=True)
    if Authentication.is_valid_token(login_token):
        Authentication.set_token(login_token)
        click.echo(f"Valid API Token saved in {SETTINGS.get_path_to_settings()}")
    else:
        click.echo("entered token is not valid. Please try again.")
        sys.exit(1)


@click.command()
def logout() -> None:
    """
    Remove all authentication tokens from global settings file
    """
    Authentication.delete_token()
    click.echo("logged out")


class Authentication:
    SEMGREP_LOGIN_TOKEN_ENVVAR_NAME = "SEMGREP_LOGIN_TOKEN"
    SEMGREP_API_TOKEN_SETTINGS_KEY = "api_token"

    @staticmethod
    def is_valid_token(token: str) -> bool:
        """
        Returns true if token is valid
        """
        import requests

        headers = {"User-Agent": SEMGREP_USER_AGENT, "Authorization": f"Bearer {token}"}
        r = requests.get(
            f"{SEMGREP_URL}api/agent/deployment", timeout=10, headers=headers
        )
        return r.ok

    @staticmethod
    def get_deployment_id() -> Optional[int]:
        """
        Returns the deployment_id attached to an api_token as int

        Returns None if api_token is invalid/doesn't have associated deployment
        """
        import requests

        token = Authentication.get_token()

        headers = {"User-Agent": SEMGREP_USER_AGENT, "Authorization": f"Bearer {token}"}
        r = requests.get(
            f"{SEMGREP_URL}api/agent/deployment", timeout=10, headers=headers
        )
        data = r.json()
        return data.get("deployment", {}).get("id")  # type: ignore

    @staticmethod
    def get_token() -> Optional[str]:
        """
        Get saved token in following order:
        - env var
        - settings file
        - None
        """
        login_token = os.environ.get(Authentication.SEMGREP_LOGIN_TOKEN_ENVVAR_NAME)
        if login_token:
            logger.debug(
                f"Using environment variable {Authentication.SEMGREP_LOGIN_TOKEN_ENVVAR_NAME} as api token"
            )
            return login_token

        return Authentication.read_token()

    @staticmethod
    def read_token() -> Optional[str]:
        """
        Read api token from settings file

        Returns None if api token not in settings file
        """
        logger.debug("Getting API token from settings file")
        login_token = SETTINGS.get_setting(
            Authentication.SEMGREP_API_TOKEN_SETTINGS_KEY, default=None
        )

        if login_token and not isinstance(login_token, str):
            raise ValueError()

        return login_token

    @staticmethod
    def set_token(token: str) -> None:
        """
        Save api token to settings file
        """
        logger.debug("Saving API token in settings file")
        SETTINGS.add_setting(Authentication.SEMGREP_API_TOKEN_SETTINGS_KEY, token)

    @staticmethod
    def delete_token() -> None:
        """
        Remove api token from settings file
        """
        logger.debug("Deleting api token from settings file")
        SETTINGS.delete_setting(Authentication.SEMGREP_API_TOKEN_SETTINGS_KEY)



import json
from pathlib import Path
from typing import Collection
from typing import Mapping
from typing import NewType
from typing import Optional

import attr

from semgrep.error import UnknownLanguageError
from semgrep.rule_lang import Span
from semgrep.types import JsonObject

Mode = NewType("Mode", str)
FileExtension = NewType("FileExtension", str)
Language = NewType("Language", str)
Shebang = str

JOIN_MODE = Mode("join")
SEARCH_MODE = DEFAULT_MODE = Mode("search")


@attr.s(auto_attribs=True, frozen=True)
class LanguageDefinition:
    """
    Mirrors schema of lang.json (see lang/README.md) for each language
    """

    id: Language
    name: str
    keys: Collection[str]
    exts: Collection[FileExtension]
    reverse_exts: Collection[str]
    shebangs: Collection[Shebang]

    @classmethod
    def from_dict(cls, data: JsonObject) -> "LanguageDefinition":
        return cls(
            id=Language(data["id"]),
            name=data["name"],
            keys=data["keys"],
            exts=data["exts"],
            reverse_exts=data.get("reverse_exts", data["exts"]),
            shebangs=data.get("shebangs", []),
        )


class _LanguageData:
    def __init__(self) -> None:
        with (Path(__file__).parent / "lang" / "lang.json").open() as fd:
            data = json.load(fd)

        self.definition_by_id: Mapping[Language, LanguageDefinition] = {
            Language(d["id"]): LanguageDefinition.from_dict(d) for d in data
        }
        self.lang_by_key: Mapping[str, Language] = {
            key.lower(): lang
            for lang, definition in self.definition_by_id.items()
            for key in definition.keys
        }
        self.lang_by_ext: Mapping[str, Language] = {
            ext: lang
            for lang, definition in self.definition_by_id.items()
            for ext in definition.reverse_exts
        }
        self.all_language_keys: Collection[str] = sorted(self.lang_by_key.keys())

    def resolve(self, lang_str: str, span: Optional[Span] = None) -> Language:
        """
        Convert an inputted string representing a language to a Language

        :param lang_str: string representing a language (e.g. "C#")
        :param span: span of language string in the original file (for error reporting),
                    None if resolve was called within semgrep
        """
        normalized = lang_str.lower()
        if normalized in self.lang_by_key:
            return self.lang_by_key[normalized]
        else:
            spans = [span.with_context(before=1, after=1)] if span else []
            raise UnknownLanguageError(
                short_msg=f"invalid language: {normalized}",
                long_msg=f"unsupported language: {normalized}. supported languages are: {', '.join(self.all_language_keys)}",
                spans=spans,
            )


LANGUAGE = _LanguageData()


ALLOWED_GLOB_TYPES = ("include", "exclude")



def test_dump_ast(run_semgrep_in_tmp, snapshot):
    stdout, _ = run_semgrep_in_tmp(
        "rules/eqeq.yaml",
        target_name="basic/stupid.py",
        options=["--dump-ast", "--lang", "python"],
    )
    snapshot.assert_match(stdout, "results.json")


def test_dump_ast_no_lang(run_semgrep_in_tmp, snapshot):
    _, stderr = run_semgrep_in_tmp(
        "rules/eqeq.yaml",
        target_name="basic/stupid.py",
        options=["--dump-ast", "python"],
        fail_on_nonzero=False,
    )
    snapshot.assert_match(stderr, "error.txt")



import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Optional

import pytest

SENTINEL_1 = 23478921

# def test_empty_commit_baseline(git_tmp_path, snapshot):
#     # Test that having a first commit that is empty (git commit --allow-empty)
#     # doesn't break semgrep
#     subprocess.run(["git", "commit", "--allow-empty", "-m", "first"], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
#     base_commit = subprocess.check_output(
#         ["git", "rev-parse", "HEAD"], encoding="utf-8"
#     ).strip()

#     # Add head finding
#     foo = git_tmp_path / "foo.py"
#     bar = git_tmp_path / "bar.py"
#     foo.write_text(f"x = {SENTINEL_1}\n")
#     bar.write_text(f"y = {SENTINEL_1}\n")
#     subprocess.run(["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
#     subprocess.run(["git", "commit", "-m", "second"], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

#     # Non-baseline scan should report findings
#     output = subprocess.run([sys.executable, "-m", "semgrep", "-e", f"$X = {SENTINEL_1}", "-l", "python"], stderr=subprocess.PIPE, stdout=subprocess.PIPE, encoding="utf-8", check=True)
#     snapshot.assert_match(output.stdout, "output.txt")
#     assert output.stdout != ""
#     snapshot.assert_match(output.stderr.replace(base_commit, "baseline-commit"), "error.txt")

#     # Baseline scan should report same findings
#     baseline_output = subprocess.run([sys.executable, "-m", "semgrep", "-e", f"$X = {SENTINEL_1}", "-l", "python", "--baseline-commit", base_commit], stderr=subprocess.PIPE, stdout=subprocess.PIPE, encoding="utf-8")
#     snapshot.assert_match(baseline_output.stdout, "baseline_output.txt")
#     # assert baseline_output.stdout == output.stdout
#     snapshot.assert_match(baseline_output.stderr.replace(base_commit, "baseline-commit"), "baseline_error.txt")
def run_sentinel_scan(check: bool = True, base_commit: Optional[str] = None):
    env = {"LANG": "en_US.UTF-8"}
    env["SEMGREP_USER_AGENT_APPEND"] = "testing"
    unique_settings_file = tempfile.NamedTemporaryFile().name
    Path(unique_settings_file).write_text("has_shown_metrics_notification: true")
    env["SEMGREP_SETTINGS_FILE"] = unique_settings_file

    cmd = [
        sys.executable,
        "-m",
        "semgrep",
        "--disable-version-check",
        "--metrics",
        "off",
        "-e",
        f"$X = {SENTINEL_1}",
        "-l",
        "python",
    ]
    if base_commit:
        cmd.extend(["--baseline-commit", base_commit])

    return subprocess.run(
        cmd,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        encoding="utf-8",
        check=check,
        env=env,
    )


def test_one_commit_with_baseline(git_tmp_path, snapshot):
    # Test that head having no change to base (git commit --allow-empty)
    # doesnt break semgrep
    foo = git_tmp_path / "foo.py"
    foo.write_text(f"x = {SENTINEL_1}\n")
    bar = git_tmp_path / "bar.py"
    bar.write_text(f"y = {SENTINEL_1}\n")

    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "first"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    base_commit = subprocess.check_output(
        ["git", "rev-parse", "HEAD"], encoding="utf-8"
    ).strip()

    # Add and commit noop change
    subprocess.run(
        ["git", "commit", "--allow-empty", "-m", "second"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    # Non-baseline scan should report findings
    output = run_sentinel_scan()
    snapshot.assert_match(output.stdout, "output.txt")
    assert (
        output.stdout != ""
    )  # If you fail this assertion means above snapshot was incorrectly changed
    snapshot.assert_match(output.stderr, "error.txt")

    # Baseline scan should report 0 findings
    baseline_output = run_sentinel_scan(base_commit=base_commit)
    assert baseline_output.stdout == ""
    snapshot.assert_match(
        baseline_output.stderr.replace(base_commit, "baseline-commit"),
        "baseline_error.txt",
    )


def test_symlink(git_tmp_path, snapshot):
    pass


def test_no_findings_both(git_tmp_path, snapshot):
    # Test if no findings in head or base semgrep doesnt explode
    foo = git_tmp_path / "foo.py"
    bar = git_tmp_path / "bar.py"
    bar.write_text(f"y = 1\n")
    foo.write_text(f"x = 1\n")

    # Add files with no finding
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "first"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    base_commit = subprocess.check_output(
        ["git", "rev-parse", "HEAD"], encoding="utf-8"
    ).strip()

    # Add files with no finding
    baz = git_tmp_path / "baz.py"
    baz.write_text("z = 1")
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "second"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    # Non-baseline scan should report no findings
    output = run_sentinel_scan()
    assert output.stdout == ""
    snapshot.assert_match(
        output.stderr.replace(base_commit, "baseline-commit"), "error.txt"
    )

    # Baseline scan should report no findings
    baseline_output = run_sentinel_scan(base_commit=base_commit)
    assert baseline_output.stdout == output.stdout
    snapshot.assert_match(
        baseline_output.stderr.replace(base_commit, "baseline-commit"),
        "baseline_error.txt",
    )


def test_no_findings_head(git_tmp_path, snapshot):
    # Test that no findings in head reports no findings even if
    # findings in baseline
    foo = git_tmp_path / "foo.py"
    bar = git_tmp_path / "bar.py"
    bar.write_text(f"y = {SENTINEL_1}\n")
    foo.write_text(f"x = {SENTINEL_1}\n")

    # Add baseline finding
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "first"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    base_commit = subprocess.check_output(
        ["git", "rev-parse", "HEAD"], encoding="utf-8"
    ).strip()

    # Add head finding
    baz = git_tmp_path / "baz.py"
    baz.write_text("z  = 1")
    foo.write_text("")  # Overwrite foo and baz to empty file
    bar.write_text("")
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "second"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    # Non-baseline scan should report no findings
    output = run_sentinel_scan()
    assert output.stdout == ""
    snapshot.assert_match(
        output.stderr.replace(base_commit, "baseline-commit"), "error.txt"
    )

    # Baseline scan should report no findings
    baseline_output = run_sentinel_scan(base_commit=base_commit)
    assert baseline_output.stdout == output.stdout
    snapshot.assert_match(
        baseline_output.stderr.replace(base_commit, "baseline-commit"),
        "baseline_error.txt",
    )


def test_no_findings_baseline(git_tmp_path, snapshot):
    # Test when head contains all findings and baseline doesnt contain any
    foo = git_tmp_path / "foo.py"
    foo.write_text(f"x = 1")

    # Add baseline finding
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "first"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    base_commit = subprocess.check_output(
        ["git", "rev-parse", "HEAD"], encoding="utf-8"
    ).strip()

    # Add head finding
    bar = git_tmp_path / "bar.py"
    bar.write_text(f"y = {SENTINEL_1}\n")
    foo.write_text(f"x = {SENTINEL_1}\n")
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "second"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    # Non-baseline scan should report findings
    output = run_sentinel_scan()
    snapshot.assert_match(output.stdout, "output.txt")
    assert output.stdout != ""
    snapshot.assert_match(
        output.stderr.replace(base_commit, "baseline-commit"), "error.txt"
    )

    # Baseline scan should report same findings
    baseline_output = run_sentinel_scan(base_commit=base_commit)
    snapshot.assert_match(baseline_output.stdout, "baseline_output.txt")
    assert baseline_output.stdout == output.stdout
    snapshot.assert_match(
        baseline_output.stderr.replace(base_commit, "baseline-commit"),
        "baseline_error.txt",
    )


def test_some_intersection(git_tmp_path, snapshot):
    # Test when baseline contains some findings of head
    foo = git_tmp_path / "foo.py"
    foo.write_text(f"x = {SENTINEL_1}\n")

    # Add baseline finding
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "first"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    base_commit = subprocess.check_output(
        ["git", "rev-parse", "HEAD"], encoding="utf-8"
    ).strip()

    # Add head finding
    bar = git_tmp_path / "bar.py"
    bar.write_text(f"y = {SENTINEL_1}\n")
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "second"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    # Non-baseline scan should report 2 findings
    output = run_sentinel_scan()
    snapshot.assert_match(output.stdout, "output.txt")
    assert output.stdout != ""
    snapshot.assert_match(
        output.stderr.replace(base_commit, "baseline-commit"), "error.txt"
    )

    # Baseline scan should report 1 finding but hide 1
    baseline_output = run_sentinel_scan(base_commit=base_commit)
    snapshot.assert_match(baseline_output.stdout, "baseline_output.txt")
    assert baseline_output.stdout != output.stdout
    snapshot.assert_match(
        baseline_output.stderr.replace(base_commit, "baseline-commit"),
        "baseline_error.txt",
    )


def test_all_intersect(git_tmp_path, snapshot):
    # Test when baseline and head contain same findings none are reported
    foo = git_tmp_path / "foo.py"
    foo.write_text(f"x = {SENTINEL_1}\n")

    bar = git_tmp_path / "bar.py"
    bar.write_text(f"y = {SENTINEL_1}\n")

    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "first"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    base_commit = subprocess.check_output(
        ["git", "rev-parse", "HEAD"], encoding="utf-8"
    ).strip()

    # Add and commit noop change
    foo.write_text(foo.read_text() + "z = 1\n")  # Note write_text overwrites
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "second"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    # Non-baseline scan should report findings
    output = run_sentinel_scan()
    snapshot.assert_match(output.stdout, "output.txt")
    assert (
        output.stdout != ""
    )  # If you fail this assertion means above snapshot was incorreclty changed
    snapshot.assert_match(output.stderr, "error.txt")

    # Baseline scan should report 0 findings
    baseline_output = run_sentinel_scan(base_commit=base_commit)
    assert baseline_output.stdout == ""
    snapshot.assert_match(
        baseline_output.stderr.replace(base_commit, "baseline-commit"),
        "baseline_error.txt",
    )


def test_no_intersection(git_tmp_path, snapshot):
    # If no intersection of baseline and head finding should still report head finding
    foo = git_tmp_path / "foo.py"
    foo.write_text(f"x = {SENTINEL_1}")

    # Add baseline finding
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "first"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    base_commit = subprocess.check_output(
        ["git", "rev-parse", "HEAD"], encoding="utf-8"
    ).strip()

    # Add head finding remove baseline finding
    bar = git_tmp_path / "bar.py"
    bar.write_text(f"y = {SENTINEL_1}\n")
    foo.write_text("")  # Overwrite foo
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "second"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    # Non-baseline scan should report 1 finding
    output = run_sentinel_scan()
    snapshot.assert_match(output.stdout, "output.txt")
    assert output.stdout != ""
    snapshot.assert_match(
        output.stderr.replace(base_commit, "baseline-commit"), "error.txt"
    )

    # Baseline scan should report same finding
    baseline_output = run_sentinel_scan(base_commit=base_commit)
    assert baseline_output.stdout == output.stdout
    snapshot.assert_match(
        baseline_output.stderr.replace(base_commit, "baseline-commit"),
        "baseline_error.txt",
    )


def test_multiple_on_same_line(git_tmp_path, snapshot):
    pass


def test_run_in_subdirectory(git_tmp_path, snapshot):
    pass


def test_unstaged_changes(git_tmp_path, snapshot):
    # Should abort if have unstaged changes
    foo = git_tmp_path / "foo"
    foo.mkdir()
    foo_a = foo / "a.py"
    foo_a.touch()
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "first"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    base_commit = subprocess.check_output(
        ["git", "rev-parse", "HEAD"], encoding="utf-8"
    ).strip()

    foo_a.write_text(f"y = {SENTINEL_1}\n")
    output = run_sentinel_scan(base_commit=base_commit, check=False)
    assert output.returncode != 0
    snapshot.assert_match(output.stderr, "error.txt")


def test_baseline_has_head_untracked(git_tmp_path, snapshot):
    pass


def test_not_git_directory(monkeypatch, tmp_path, snapshot):
    # Should abort baseline scan if not a git directory
    monkeypatch.chdir(tmp_path)
    foo = tmp_path / "foo"
    foo.mkdir()
    foo_a = foo / "a.py"
    foo_a.write_text("y = 55555555\n")

    output = run_sentinel_scan(base_commit="12345", check=False)
    assert output.returncode != 0
    snapshot.assert_match(output.stderr, "error.txt")


def test_commit_doesnt_exist(git_tmp_path, snapshot):
    # Should abort baseline scan if baseline is not valid commit
    foo = git_tmp_path / "foo"
    foo.mkdir()
    foo_a = foo / "a.py"
    foo_a.touch()
    subprocess.run(
        ["git", "add", "."], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "commit", "-m", "first"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    output = run_sentinel_scan(base_commit="12345", check=False)
    assert output.returncode != 0
    snapshot.assert_match(output.stderr, "error.txt")


@pytest.fixture
def git_tmp_path(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    # Initialize State
    subprocess.run(
        ["git", "init"], check=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    subprocess.run(
        ["git", "config", "user.email", "baselinetest@r2c.dev"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    subprocess.run(
        ["git", "config", "user.name", "Baseline Test"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    subprocess.run(
        ["git", "checkout", "-B", "main"],
        check=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    yield tmp_path



import pytest


@pytest.mark.parametrize("max_bytes", ["1MB", "1B", "1.3R"])
def test_max_target_bytes(run_semgrep_in_tmp, snapshot, max_bytes):
    stdout, stderr = run_semgrep_in_tmp(
        "rules/eqeq.yaml",
        options=["--max-target-bytes", max_bytes],
        fail_on_nonzero=False,
    )
    snapshot.assert_match(stdout, "results.json")
    snapshot.assert_match(stderr, "error.txt")



import json
import subprocess
import tempfile
from pathlib import Path
from subprocess import CalledProcessError

import pytest
from tests.conftest import _clean_output_json
from tests.conftest import _mask_times

from semgrep.constants import OutputFormat

GITHUB_TEST_GIST_URL = (
    "https://raw.githubusercontent.com/returntocorp/semgrep-rules/develop/template.yaml"
)


def test_basic_rule__local(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(run_semgrep_in_tmp("rules/eqeq.yaml")[0], "results.json")


def test_basic_rule__relative(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/../rules/eqeq.yaml")[0],
        "results.json",
    )


def test_deduplication(run_semgrep_in_tmp, snapshot):
    """
    Check that semgrep runs a rule only once even when different in the metadata
    """
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/duplicate-rule.yaml", target_name="basic/stupid.py")[
            0
        ],
        "results.json",
    )


def test_noextension_filtering(run_semgrep_in_tmp, snapshot):
    """
    Check that semgrep does not filter out files without extensions when
    said file is explicitly passed
    """
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/eqeq-python.yaml", target_name="basic/stupid_no_extension"
        )[0],
        "results.json",
    )


def test_noextension_filtering_optimizations(run_semgrep_in_tmp, snapshot):
    """
    Check that semgrep does not filter out files without extensions when
    said file is explicitly passed
    """
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/eqeq-python.yaml",
            target_name="basic/stupid_no_extension",
            options=["--optimizations", "all"],
        )[0],
        "results.json",
    )


def test_script(run_semgrep_in_tmp, snapshot):
    """
    Validates that Semgrep scans scripts with matching shebangs
    """
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/eqeq-python.yaml",
            target_name="script/",
        )[0],
        "results.json",
    )


def test_basic_rule__absolute(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(Path.cwd() / "rules" / "eqeq.yaml")[0],
        "results.json",
    )


def test_terminal_output(run_semgrep_in_tmp, snapshot):
    # Have shared settings file to test second run doesnt show metric output
    settings_file = tempfile.NamedTemporaryFile().name

    text_output = run_semgrep_in_tmp(
        "rules/eqeq.yaml", output_format=OutputFormat.TEXT, settings_file=settings_file
    )
    snapshot.assert_match(text_output[0], "output.txt")
    snapshot.assert_match(text_output[1], "error.txt")

    # Metric message should not appear in second output
    text_output = run_semgrep_in_tmp(
        "rules/eqeq.yaml", output_format=OutputFormat.TEXT, settings_file=settings_file
    )
    snapshot.assert_match(text_output[0], "output_second.txt")
    snapshot.assert_match(text_output[1], "error_second.txt")


def test_terminal_output_quiet(run_semgrep_in_tmp, snapshot):
    """
    Quiet output should just have finding output
    """
    text_output = run_semgrep_in_tmp(
        "rules/eqeq.yaml",
        output_format=OutputFormat.TEXT,
        quiet=True,
        # Pass named temporary file to force metric notice behavior on first scan
        # (but should not see anything cause of --quiet)
        settings_file=tempfile.NamedTemporaryFile().name,
    )
    snapshot.assert_match(text_output[0], "output.txt")
    snapshot.assert_match(text_output[1], "error.txt")


def test_stdin_input(snapshot):
    process = subprocess.Popen(
        [
            "python3",
            "-m",
            "semgrep",
            "--disable-version-check",
            "--metrics",
            "off",
            "--json",
            "-e",
            "a",
            "--lang",
            "js",
            "-",
        ],
        encoding="utf-8",
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    stdout, _ = process.communicate("a")
    snapshot.assert_match(_clean_output_json(stdout), "results.json")


def test_subshell_input(snapshot):
    stdout = subprocess.check_output(
        [
            "bash",
            "-c",
            "python3 -m semgrep --disable-version-check --metrics off --json -e 'a' --lang js <(echo 'a')",
        ],
        encoding="utf-8",
    )
    snapshot.assert_match(_clean_output_json(stdout), "results.json")


def test_multi_subshell_input(snapshot):
    stdout = subprocess.check_output(
        [
            "bash",
            "-c",
            "python3 -m semgrep --disable-version-check --metrics off --json -e 'a' --lang js <(echo 'a') <(echo 'b + a')",
        ],
        encoding="utf-8",
    )
    snapshot.assert_match(_clean_output_json(stdout), "results.json")


def test_multiline(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/eqeq.yaml", target_name="multiline")[0],
        "results.json",
    )


def test_url_rule(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(GITHUB_TEST_GIST_URL)[0],
        "results.json",
    )


def test_registry_rule(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("r2c")[0],
        "results.json",
    )


def test_auto_config(run_semgrep_in_tmp):
    # --config auto will change over time, so lets just make sure this doesn't error out
    # TODO: Mock config response for more detailed testing
    # Use --no-strict to avoid error from unmatched nosem comment
    run_semgrep_in_tmp("auto", strict=False)
    assert True


def test_hidden_rule__explicit(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(run_semgrep_in_tmp("rules/hidden/.hidden")[0], "results.json")


def test_hidden_rule__implicit(run_semgrep_in_tmp, snapshot):
    with pytest.raises(CalledProcessError) as excinfo:
        run_semgrep_in_tmp("rules/hidden")[0]
    assert excinfo.value.returncode == 7
    snapshot.assert_match(excinfo.value.stdout, "error.json")

    with pytest.raises(CalledProcessError) as excinfo:
        run_semgrep_in_tmp("rules/hidden", output_format=OutputFormat.TEXT)[0]
    assert excinfo.value.returncode == 7
    snapshot.assert_match(excinfo.value.stderr, "error.txt")


def test_default_rule__file(run_semgrep_in_tmp, snapshot):
    Path(".semgrep.yml").symlink_to(Path("rules/eqeq.yaml").resolve())
    snapshot.assert_match(run_semgrep_in_tmp()[0], "results.json")


def test_default_rule__folder(run_semgrep_in_tmp, snapshot):
    Path(".semgrep").mkdir()
    Path(".semgrep/.semgrep.yml").symlink_to(Path("rules/eqeq.yaml").resolve())

    snapshot.assert_match(run_semgrep_in_tmp()[0], "results.json")


def test_regex_rule__top(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(run_semgrep_in_tmp("rules/regex-top.yaml")[0], "results.json")


def test_regex_rule__utf8(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/regex-utf8.yaml", target_name="basic/regex-utf8.txt")[
            0
        ],
        "results.json",
    )


def test_regex_rule__utf8_on_image(run_semgrep_in_tmp, snapshot):
    # https://github.com/returntocorp/semgrep/issues/4258
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/regex-utf8.yaml", target_name="image/semgrep.png")[0],
        "results.json",
    )


def test_regex_rule__child(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/regex-child.yaml")[0], "results.json"
    )


def test_regex_rule__not(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/pattern-not-regex/regex-not.yaml", target_name="basic/stupid.py"
        )[0],
        "results.json",
    )


def test_regex_rule__not2(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/pattern-not-regex/regex-not2.yaml",
            target_name="basic/regex-any-language.html",
        )[0],
        "results.json",
    )


def test_regex_rule__pattern_regex_and_pattern_not_regex(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/pattern-not-regex/regex-not-with-pattern-regex.yaml",
            target_name="basic/regex-any-language.html",
        )[0],
        "results.json",
    )


def test_regex_rule__issue2465(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/pattern-not-regex/issue2465.yaml",
            target_name="pattern-not-regex/issue2465.requirements.txt",
        )[0],
        "results.json",
    )


def test_regex_rule__invalid_expression(run_semgrep_in_tmp, snapshot):
    with pytest.raises(CalledProcessError) as excinfo:
        run_semgrep_in_tmp("rules/regex-invalid.yaml")[0]
    assert excinfo.value.returncode == 2
    snapshot.assert_match(excinfo.value.stderr, "error.txt")
    snapshot.assert_match(excinfo.value.stdout, "error.json")


def test_nested_patterns_rule(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/nested-patterns.yaml")[0], "results.json"
    )


def test_nested_pattern_either_rule(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/nested-pattern-either.yaml")[0], "results.json"
    )


def test_metavariable_regex_rule(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/metavariable-regex.yaml")[0], "results.json"
    )


def test_metavariable_regex_multi_rule(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/metavariable-regex-multi-rule.yaml")[0],
        "results.json",
    )


def test_metavariable_multi_regex_rule(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/metavariable-regex-multi-regex.yaml")[0],
        "results.json",
    )


def test_regex_with_any_language_rule(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/regex-any-language.yaml", target_name="basic/regex-any-language.html"
        )[0],
        "results.json",
    )


def test_regex_with_any_language_multiple_rule(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/regex-any-language-multiple.yaml",
            target_name="basic/regex-any-language.html",
        )[0],
        "results.json",
    )


def test_invalid_regex_with_any_language_rule(run_semgrep_in_tmp, snapshot):
    with pytest.raises(CalledProcessError) as excinfo:
        run_semgrep_in_tmp(
            "rules/regex-any-language-invalid.yaml",
            target_name="basic/regex-any-language.html",
        )
    assert excinfo.value.returncode not in (0, 1)
    snapshot.assert_match(excinfo.value.stderr, "error.txt")
    snapshot.assert_match(excinfo.value.stdout, "error.json")


def test_regex_with_any_language_rule_none_alias(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/regex-any-language-alias-none.yaml",
            target_name="basic/regex-any-language.html",
        )[0],
        "results.json",
    )


def test_regex_with_any_language_multiple_rule_none_alias(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/regex-any-language-multiple-alias-none.yaml",
            target_name="basic/regex-any-language.html",
        )[0],
        "results.json",
    )


def test_timeout(run_semgrep_in_tmp, snapshot):
    # Check that semgrep-core timeouts are properly handled

    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/long.yaml",
            options=["--timeout", "1"],
            target_name="equivalence",
            strict=False,
        )[0],
        "results.json",
    )


def test_spacegrep_timeout(run_semgrep_in_tmp, snapshot):
    # Check that spacegrep timeouts are handled gracefully.
    #
    # The pattern is designed to defeat any optimization that would
    # prevent a timeout. Both the words 'Frob' and 'Yoyodyne' occur
    # once in the file but in a different order, preventing any match.
    #
    pattern = "$A ... $B ... $C ... Frob ... Yoyodyne"

    stdout, stderr = run_semgrep_in_tmp(
        config=None,
        target_name="spacegrep_timeout/gnu-lgplv2.txt",
        options=[
            "--lang",
            "generic",
            "--pattern",
            pattern,
            "--timeout",
            "1",
        ],
        output_format=OutputFormat.TEXT,
        strict=False,  # don't fail due to timeout
    )

    snapshot.assert_match(stdout, "results.json")
    snapshot.assert_match(stderr, "error.txt")


def test_max_memory(run_semgrep_in_tmp, snapshot):
    stdout, stderr = run_semgrep_in_tmp(
        "rules/long.yaml",
        options=["--verbose", "--max-memory", "1"],
        target_name="equivalence",
        strict=False,
    )
    snapshot.assert_match(_mask_times(stdout), "results.json")
    snapshot.assert_match(stderr, "error.txt")


def test_stack_size(run_semgrep_in_tmp, snapshot):
    """
    Verify that semgrep raises the soft stack limit if possible
    when calling semgrep core
    """

    # long.yaml and equivalence were chosen since they happen to cause
    # stack exhaustion
    e2e_dir = Path(__file__).parent
    targetpath = Path(e2e_dir / "targets").resolve() / "equivalence"
    rulepath = Path(e2e_dir / "rules").resolve() / "long.yaml"

    # Set the hard as well as the soft stack limit. This should force a stack
    # overflow. If this fails, the test is broken and needs to be fixed.
    # Do not just delete this assertion. It means the actual test below does
    # not accurately verify that we are solving the stack exhaustion
    output = subprocess.run(
        f"ulimit -s 1000 && semgrep --disable-version-check --metrics off --config {rulepath} --verbose {targetpath}",
        shell=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        encoding="utf-8",
    )
    print(output.stderr)
    assert (
        "semgrep-core exit code: -11" in output.stderr
        or "Stack overflow" in output.stderr
    )

    # If only set soft limit, semgrep should raise it as necessary so we don't hit soft limit
    output = subprocess.run(
        f"ulimit -S -s 1000 && semgrep --disable-version-check --metrics off --config {rulepath} --verbose {targetpath}",
        shell=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        encoding="utf-8",
    )
    assert "semgrep-core exit code: -11" not in output.stderr
    assert "Stack overflow" not in output.stderr


def test_timeout_threshold(run_semgrep_in_tmp, snapshot):
    results = run_semgrep_in_tmp(
        "rules/multiple-long.yaml",
        options=["--verbose", "--timeout", "1", "--timeout-threshold", "1"],
        target_name="equivalence",
        strict=False,
    )[0]
    snapshot.assert_match(
        _mask_times(results),
        "results.json",
    )

    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/multiple-long.yaml",
            output_format=OutputFormat.TEXT,
            options=["--verbose", "--timeout", "1", "--timeout-threshold", "1"],
            target_name="equivalence",
            strict=False,
        )[1],
        "error.txt",
    )

    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/multiple-long.yaml",
            output_format=OutputFormat.TEXT,
            options=["--verbose", "--timeout", "1", "--timeout-threshold", "2"],
            target_name="equivalence",
            strict=False,
        )[1],
        "error_2.txt",
    )


def test_metavariable_comparison_rule(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/metavariable-comparison.yaml")[0], "results.json"
    )


def test_metavariable_comparison_rule_base(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/metavariable-comparison-base.yaml")[0], "results.json"
    )


def test_metavariable_comparison_rule_strip(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/metavariable-comparison-strip.yaml")[0],
        "results.json",
    )


def test_metavariable_comparison_rule_bad_content(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/metavariable-comparison-bad-content.yaml")[0],
        "results.json",
    )


def test_multiple_configs_file(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(["rules/eqeq.yaml", "rules/eqeq-python.yaml"])[0],
        "results.json",
    )


def test_multiple_configs_different_origins(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(["rules/eqeq.yaml", GITHUB_TEST_GIST_URL])[0], "results.json"
    )


def test_metavariable_propagation_regex(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/metavariable_propagation/metavariable-regex-propagation.yaml",
            target_name="metavariable_propagation/metavariable-regex-propagation.py",
        )[0],
        "results.json",
    )


def test_metavariable_propagation_comparison(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/metavariable_propagation/metavariable-comparison-propagation.yaml",
            target_name="metavariable_propagation/metavariable-comparison-propagation.py",
        )[0],
        "results.json",
    )


def test_taint_mode(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/taint.yaml",
            target_name="taint/taint.py",
        )[0],
        "results.json",
    )


def test_deduplication_same_message(run_semgrep_in_tmp, snapshot):
    """
    With same message, should deduplicate and only have one finding
    """
    output, _ = run_semgrep_in_tmp(
        "rules/deduplication/duplication-same-message.yaml",
        target_name="deduplication/deduplication.py",
    )
    snapshot.assert_match(output, "results.json")
    json_output = json.loads(output)
    assert len(json_output["results"]) == 1


def test_deduplication_different_message(run_semgrep_in_tmp, snapshot):
    output, _ = run_semgrep_in_tmp(
        "rules/deduplication/duplication-different-message.yaml",
        target_name="deduplication/deduplication.py",
    )
    snapshot.assert_match(output, "results.json")
    json_output = json.loads(output)
    assert len(json_output["results"]) == 2


def test_pattern_regex_empty_file(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "rules/pattern-regex-empty-file.yaml",
            target_name="empty/totally_empty_file",
        )[0],
        "results.json",
    )


def test_cdn_ruleset_resolution(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("p/ci")[0],
        "results.json",
    )



import re
from pathlib import Path
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

from semgrep.error import SemgrepError
from semgrep.rule_match import RuleMatch
from semgrep.rule_match_map import RuleMatchMap
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)

SPLIT_CHAR = "\n"


class Fix:
    def __init__(self, fixed_contents: str, fixed_lines: List[str]):
        self.fixed_contents = fixed_contents
        self.fixed_lines = fixed_lines


class FileOffsets:
    """
    This object is used to track state when applying multiple fixes to the same
    or subsequent lines in a single file.

    The assumption and current state here is that fixes are applied top-to-
    bottom and in a single pass; semgrep will not come back and re-parse or
    re-fix a file or line. This approach may need to be revisited or extended
    to support more complex overlapping autofix cases in future.
    """

    def __init__(self, line_offset: int, col_offset: int, active_line: int):
        self.line_offset = line_offset
        self.col_offset = col_offset
        self.active_line = active_line


def _get_lines(path: Path) -> List[str]:
    contents = path.read_text()
    lines = contents.split(SPLIT_CHAR)
    return lines


def _get_match_context(
    rule_match: RuleMatch, offsets: FileOffsets
) -> Tuple[int, int, int, int]:
    start_obj = rule_match.start
    start_line = start_obj.line - 1  # start_line is 1 indexed
    start_col = start_obj.col - 1  # start_col is 1 indexed
    end_obj = rule_match.end
    end_line = end_obj.line - 1  # end_line is 1 indexed
    end_col = end_obj.col - 1  # end_line is 1 indexed

    # adjust based on offsets
    start_line = start_line + offsets.line_offset
    end_line = end_line + offsets.line_offset
    start_col = start_col + offsets.col_offset
    end_col = end_col + offsets.col_offset

    return start_line, start_col, end_line, end_col


def _basic_fix(
    rule_match: RuleMatch, file_offsets: FileOffsets, fix: str
) -> Tuple[Fix, FileOffsets]:
    p = rule_match.path
    lines = _get_lines(p)

    # get the start and end points
    start_line, start_col, end_line, end_col = _get_match_context(
        rule_match, file_offsets
    )

    # break into before, to modify, after
    before_lines = lines[:start_line]
    before_on_start_line = lines[start_line][:start_col]
    after_on_end_line = lines[end_line][end_col:]  # next char after end of match
    modified_lines = (before_on_start_line + fix + after_on_end_line).splitlines()
    after_lines = lines[end_line + 1 :]  # next line after end of match
    contents_after_fix = before_lines + modified_lines + after_lines

    # update offsets
    file_offsets.line_offset = len(contents_after_fix) - len(lines)
    if start_line == end_line:
        file_offsets.col_offset = len(fix) - (end_col - start_col)

    return Fix(SPLIT_CHAR.join(contents_after_fix), modified_lines), file_offsets


def _regex_replace(
    rule_match: RuleMatch,
    file_offsets: FileOffsets,
    from_str: str,
    to_str: str,
    count: int = 1,
) -> Tuple[Fix, FileOffsets]:
    """
    Use a regular expression to autofix.
    Replaces from_str to to_str, starting from the left,
    exactly `count` times.
    """
    path = rule_match.path
    lines = _get_lines(path)

    start_line, _, end_line, _ = _get_match_context(rule_match, file_offsets)

    before_lines = lines[:start_line]
    after_lines = lines[end_line + 1 :]

    match_context = lines[start_line : end_line + 1]

    fix = re.sub(from_str, to_str, "\n".join(match_context), count)
    modified_context = fix.splitlines()
    modified_contents = before_lines + modified_context + after_lines

    # update offsets
    file_offsets.line_offset = len(modified_context) - len(match_context)

    return Fix(SPLIT_CHAR.join(modified_contents), modified_context), file_offsets


def _write_contents(path: Path, contents: str) -> None:
    path.write_text(contents)


def apply_fixes(rule_matches_by_rule: RuleMatchMap, dryrun: bool = False) -> None:
    """
    Modify files in place for all files with findings from rules with an
    autofix configuration
    """
    modified_files: Set[Path] = set()
    modified_files_offsets: Dict[Path, FileOffsets] = {}
    for _, rule_matches in rule_matches_by_rule.items():
        for rule_match in rule_matches:
            fix = rule_match.fix
            fix_regex = rule_match.fix_regex
            filepath = rule_match.path
            # initialize or retrieve/update offsets for the file
            file_offsets = modified_files_offsets.get(
                filepath, FileOffsets(0, 0, rule_match.start.line)
            )
            if file_offsets.active_line != rule_match.start.line:
                file_offsets.active_line = rule_match.start.line
                file_offsets.col_offset = 0
            if fix:
                try:
                    fixobj, modified_files_offsets[filepath] = _basic_fix(
                        rule_match, file_offsets, fix
                    )
                except Exception as e:
                    raise SemgrepError(f"unable to modify file {filepath}: {e}")
            elif fix_regex:
                regex = fix_regex.get("regex")
                replacement = fix_regex.get("replacement")
                count = fix_regex.get("count", 0)
                if not regex or not replacement:
                    raise SemgrepError(
                        "'regex' and 'replacement' values required when using 'fix-regex'"
                    )
                try:
                    count = int(count)
                except ValueError:
                    raise SemgrepError(
                        "optional 'count' value must be an integer when using 'fix-regex'"
                    )
                try:
                    fixobj, modified_files_offsets[filepath] = _regex_replace(
                        rule_match, file_offsets, regex, replacement, count
                    )
                except Exception as e:
                    raise SemgrepError(
                        f"unable to use regex to modify file {filepath} with fix '{fix}': {e}"
                    )
            else:
                continue
            # endif
            if not dryrun:
                _write_contents(rule_match.path, fixobj.fixed_contents)
                modified_files.add(filepath)
            else:
                rule_match.extra[
                    "fixed_lines"
                ] = fixobj.fixed_lines  # Monkey patch in fixed lines

    num_modified = len(modified_files)
    if len(modified_files):
        logger.info(
            f"successfully modified {num_modified} file{'s' if num_modified > 1 else ''}."
        )
    else:
        logger.info(f"no files modified.")



import fnmatch
import os
import re
from pathlib import Path
from typing import Iterable
from typing import Iterator
from typing import Set
from typing import TextIO

import attr

from semgrep.error import SemgrepError
from semgrep.types import FilteredTargets
from semgrep.util import partition_set
from semgrep.verbose_logging import getLogger

CONTROL_REGEX = re.compile(r"(?!<\\):")  # Matches unescaped colons
MULTI_CHAR_REGEX = re.compile(
    r"(?!<\\)\[.*(?!<\\)\]"
)  # Matches anything in unescaped brackets
COMMENT_START_REGEX = re.compile(r"(?P<ignore_pattern>.*?)(?:\s+|^)#.*")
IGNORE_FILE_NAME = ".semgrepignore"

logger = getLogger(__name__)

# For some reason path.is_relative_to produces a complaint that 'PosixPath' object has no attribute 'is_relative_to'
# So we just copy its implementation
def path_is_relative_to(p1: Path, p2: Path) -> bool:
    try:
        p1.relative_to(p2)
        return True
    except ValueError:
        return False


## TODO: This files duplicates the .semgrepignore functionality from semgrep-action.
## We should ultimately remove this from semgrep-action, and keep it as part of the CLI

# This class is a duplicate of the FileIgnore class in semgrep-action, but with all file walking functionality removed
@attr.s
class FileIgnore:
    base_path = attr.ib(type=Path)
    patterns = attr.ib(type=Set[str])
    _processed_patterns = attr.ib(type=Set[str], init=False)

    def __attrs_post_init__(self) -> None:
        self._processed_patterns = Processor(self.base_path).process(self.patterns)

    def _survives(self, path: Path) -> bool:
        """
        Determines if a single Path survives the ignore filter.
        """
        for p in self._processed_patterns:
            if path.is_dir() and p.endswith("/") and fnmatch.fnmatch(str(path), p[:-1]):
                logger.verbose(f"Ignoring {path} due to .semgrepignore")
                return False
            if fnmatch.fnmatch(str(path), p):
                logger.verbose(f"Ignoring {path} due to .semgrepignore")
                return False

            # Check any subpath of path satisfies a pattern
            # i.e. a/b/c/d is ignored with rule a/b
            # This is a hack. TargetFileManager should be pruning while searching
            # instead of post filtering to avoid this
            # Note: Use relative to base to avoid ignore rules firing on parent directories
            # i.e. /private/var/.../instabot should not be ignored with var/ rule
            # in instabot dir as base_path
            # Note: Append "/" to path before running fnmatch so **/pattern matches with pattern/stuff
            if (
                path_is_relative_to(path, self.base_path)
                and p.endswith("/")
                and fnmatch.fnmatch(
                    "/" + str(path.relative_to(self.base_path)), p + "*"
                )
            ):
                logger.verbose(f"Ignoring {path} due to .semgrepignore")
                return False
            if (
                p.endswith("/")
                and p.startswith(str(self.base_path))
                and fnmatch.fnmatch(str(path), p + "*")
            ):
                logger.verbose(f"Ignoring {path} due to .semgrepignore")
                return False

        return True

    def filter_paths(self, *, candidates: Iterable[Path]) -> FilteredTargets:
        kept, removed = partition_set(
            lambda path: path.exists()
            and (
                self._survives(path.absolute())
                or path.absolute().samefile(self.base_path)
            ),
            candidates,
        )
        return FilteredTargets(kept, removed)


# This class is an exact duplicate of the Parser class in semgrep-action
@attr.s(auto_attribs=True)
class Parser:
    """
    A parser for semgrepignore syntax.

    semgrepignore syntax mirrors gitignore syntax, with the following modifications:
    - "Include" patterns (lines starting with "!") are not supported.
    - "Character range" patterns (lines including a collection of characters inside brackets) are not supported.
    - An ":include ..." directive is added, which allows another file to be included in the ignore pattern list;
      typically this included file would be the project .gitignore. No attempt at cycle detection is made.
    - Any line beginning with a colon, but not ":include ", will raise a SemgrepError.
    - "\:" is added to escape leading colons.

    Unsupported patterns are silently removed from the pattern list (this is done so that gitignore files may be
    included without raising errors), although the removal will be logged.

    Unfortunately there's no available parser for gitignore syntax in python, so we have
    to make our own. The syntax is simple enough that we can just roll our own parser, so
    I deliberately skip using a parser generator or combinator library, which would either need to
    parse on a character-by-character basis or make use of a large number of regex scans.

    The parser steps are, for each line in the input stream:
    1. Remove comments
    2. Remove unsupported gitignore syntax
    3. Expand directives

    The end result of this parsing is a set of human-readable patterns corresponding to gitignore syntax.
    To use these patterns with fnmatch, however, a final postprocessing step is needed, achieved by calling
    Processor.process().

    :param base_path:   The path relative to which :include directives should be evaluated
    """

    # Parser steps are each represented as Generators. This allows us to chain
    # steps, whether the step is a transformation, a filter, an expansion, or any combination thereof.

    base_path: Path

    @staticmethod
    def remove_comments(line: str) -> Iterator[str]:
        """If a line has a comment, remove the comment and just return the ignore pattern"""
        m = COMMENT_START_REGEX.match(line)
        if m:
            yield m.groupdict().get(
                "ignore_pattern", ""
            )  # return empty string if entire line is a comment
        else:
            yield line.rstrip()

    @staticmethod
    def filter_supported(line: str) -> Iterator[str]:
        """Remove unsupported gitignore patterns"""
        if not line:
            pass
        elif line.startswith("!") or MULTI_CHAR_REGEX.search(line):
            logger.debug(f"Skipping unsupported gitignore pattern '{line}'")
        else:
            yield line

    def expand_directives(self, line: str) -> Iterable[str]:
        """Load :include files"""
        if line.startswith(":include "):
            include_path = self.base_path / line[9:]
            if include_path.is_file():
                with include_path.open() as include_lines:
                    sub_base = include_path.parent.resolve()
                    sub_parser = Parser(sub_base)
                    return sub_parser.parse(include_lines)
            else:
                logger.debug(
                    f"Skipping `:include {include_path}` directive, file not found"
                )
                return []
        elif CONTROL_REGEX.match(line):
            raise SemgrepError(
                f"Unknown ignore directive in Semgrep ignore file at {self.base_path}: '{line}'"
            )
        else:
            return (line for _ in range(1))

    def parse(self, stream: TextIO) -> Set[str]:
        """Performs parsing of an input stream"""
        return {
            pattern
            for line in stream
            for no_comments in self.remove_comments(line)
            for supported in self.filter_supported(no_comments)
            for pattern in self.expand_directives(supported)
        }


# This class is an exact duplicate of the Processor class in semgrep-action
@attr.s(auto_attribs=True)
class Processor:
    """
    A post-processor for parsed semgrepignore files.

    The postprocessor is responsible for converting the parser's intermediate representation to a set of
    patterns compatible with fnmatch. The steps are:
    1. Unescape escape characters
    2. Convert gitignore patterns into fnmatch patterns
    """

    # Per Parser, each Processor step is represented as a Generator.

    base_path: Path

    @staticmethod
    def unescape(line: str) -> Iterator[str]:
        """Expands escape characters"""
        out = ""
        is_escape = False
        for c in line:
            if is_escape:
                out += c
                is_escape = False
            elif c == "\\":
                is_escape = True
            else:
                out += c
        yield out

    def to_fnmatch(self, pat: str) -> Iterator[str]:
        """Convert a single pattern from gitignore to fnmatch syntax"""
        if pat.rstrip("/").find("/") < 0:
            # Handles:
            #   file
            #   path/
            pat = os.path.join("**", pat)
        if pat.startswith("./") or pat.startswith("/"):
            # Handles:
            #   /relative/to/root
            #   ./relative/to/root
            pat = pat.lstrip(".").lstrip("/")
        if not pat.startswith("**"):
            # Handles:
            #   path/to/absolute
            #   */to/absolute
            #   path/**/absolute
            pat = os.path.join(self.base_path, pat)
        yield pat

    def process(self, pre: Iterable[str]) -> Set[str]:
        """Post-processes an intermediate representation"""
        return {
            pattern
            for pat in pre
            for unescaped in self.unescape(pat)
            for pattern in self.to_fnmatch(unescaped)
        }



from semgrep.semgrep_types import LANGUAGE


def test_no_duplicate_keys() -> None:
    """
    Ensures one-to-one assumption of mapping from keys to language in lang.json
    """
    keys = set()
    for d in LANGUAGE.definition_by_id.values():
        for k in d.keys:
            if k in keys:
                raise Exception(f"Duplicate language key {k}")
            keys.add(k)


def test_no_duplicate_reverse_exts() -> None:
    """
    Ensures one-to-one assumption of mapping from reverse file extensions to language in lang.json
    """
    exts = set()
    for d in LANGUAGE.definition_by_id.values():
        for e in d.reverse_exts:
            if e in exts:
                raise Exception(f"Duplicate language key {e}")
            exts.add(e)



"""
Manages system settings for Semgrep

These can change the way Semgrep operates within the environment. They are distinct from config,
which describes which rules Semgrep should run.

To retrieve settings use `SETTINGS.value`, to set them use `SETTINGS.value = ...`.

If no settings have been configured on the system, DEFAULT_SETTINGS will be written.

If the process does not have permission to the settings path, a PermissionError will be raised;
callers should handle this gracefully.
"""
import os
from pathlib import Path
from typing import Any
from typing import cast
from typing import Dict
from typing import Mapping

from ruamel.yaml import YAML

from semgrep.constants import SEMGREP_SETTINGS_FILE
from semgrep.constants import SETTINGS_FILE
from semgrep.constants import USER_DATA_FOLDER
from semgrep.verbose_logging import getLogger

DEFAULT_SETTINGS: Dict[str, Any] = {}


class Settings:
    def __init__(self) -> None:
        self._logger = getLogger(__name__)
        self._yaml = YAML()
        self._yaml.default_flow_style = False
        self._path = Settings.get_path_to_settings()

        # If file exists, read file. Otherwise use default
        # Must perform access check first in case we don't have permission to stat the path
        if os.access(self._path, os.R_OK) and self._path.exists():
            with self._path.open("r") as fd:
                yaml_file = self._yaml.load(fd)
            if not isinstance(yaml_file, Mapping):
                self._logger.warning(
                    f"Bad settings format; {self._path} will be overriden. Contents:\n{yaml_file}"
                )
                self._value = DEFAULT_SETTINGS
            else:
                self._value = cast(Dict[str, Any], yaml_file)
        else:
            self._value = DEFAULT_SETTINGS

    @staticmethod
    def get_path_to_settings() -> Path:
        """
        Uses ~/.semgrep/settings.yaml unless SEMGREP_SETTINGS_FILE is set
        """
        if SEMGREP_SETTINGS_FILE:
            return Path(SEMGREP_SETTINGS_FILE)

        return Path("~").expanduser() / USER_DATA_FOLDER / SETTINGS_FILE

    def add_setting(self, key: str, value: Any) -> None:
        """
        Update and save this system's settings object

        :param value: The settings object
        """
        self._value[key] = value
        self._write_settings()

    def _write_settings(self) -> None:
        try:
            if not self._path.parent.exists():
                self._path.parent.mkdir(parents=True, exist_ok=True)
            with self._path.open("w") as fd:
                self._yaml.dump(self._value, fd)
        except PermissionError:
            self._logger.verbose("Could not write settings file at %s", self._path)

    def get_setting(self, key: str, *, default: Any) -> Any:
        return self._value.get(key, default)

    def delete_setting(self, key: str) -> None:
        """
        Deletes key KEY from settings file if it exists

        Noop otherwise
        """
        if key in self._value:
            del self._value[key]
            self._write_settings()


SETTINGS = Settings()



import os
import re
from enum import auto
from enum import Enum
from typing import Type

from semgrep import __VERSION__

RULES_KEY = "rules"
ID_KEY = "id"
CLI_RULE_ID = "-"
SEMGREP_URL = os.environ.get("SEMGREP_URL", "https://semgrep.dev/")
PLEASE_FILE_ISSUE_TEXT = "An error occurred while invoking the Semgrep engine. Please help us fix this by creating an issue at https://github.com/returntocorp/semgrep"

DEFAULT_SEMGREP_CONFIG_NAME = "semgrep"
DEFAULT_CONFIG_FILE = f".{DEFAULT_SEMGREP_CONFIG_NAME}.yml"
DEFAULT_CONFIG_FOLDER = f".{DEFAULT_SEMGREP_CONFIG_NAME}"

DEFAULT_TIMEOUT = 30  # seconds

USER_DATA_FOLDER = ".semgrep"
SETTINGS_FILE = "settings.yml"
SEMGREP_SETTING_ENVVAR_NAME = "SEMGREP_SETTINGS_FILE"
SEMGREP_SETTINGS_FILE = os.environ.get(SEMGREP_SETTING_ENVVAR_NAME)

SEMGREP_USER_AGENT = f"Semgrep/{__VERSION__}"
SEMGREP_USER_AGENT_APPEND = os.environ.get("SEMGREP_USER_AGENT_APPEND")
if SEMGREP_USER_AGENT_APPEND is not None:
    SEMGREP_USER_AGENT = f"{SEMGREP_USER_AGENT} {SEMGREP_USER_AGENT_APPEND}"

SEMGREP_CDN_BASE_URL = os.environ.get("SEMGREP_CDN_BASE_URL", "https://cdn.semgrep.dev")

YML_EXTENSIONS = {".yml", ".yaml"}
YML_SUFFIXES = [[ext] for ext in YML_EXTENSIONS]
YML_TEST_SUFFIXES = [[".test", ext] for ext in YML_EXTENSIONS]


class OutputFormat(Enum):
    TEXT = auto()
    JSON = auto()
    JUNIT_XML = auto()
    SARIF = auto()
    EMACS = auto()
    VIM = auto()

    def is_json(self) -> bool:
        return self in [OutputFormat.JSON, OutputFormat.SARIF]


# Ensure consistency with 'severity' in 'rule_schema.yaml'
class RuleSeverity(Enum):
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    INVENTORY = "INVENTORY"

    @classmethod
    def _missing_(cls: Type[Enum], value: object) -> Enum:
        if not isinstance(value, str):
            raise TypeError(f"invalid rule severity type: {type(value)}")
        for member in cls:
            if member.value.lower() == value:
                return member
        raise ValueError(f"invalid rule severity value: {value}")


RULE_ID_RE_STR = r"(?:[:=][\s]?(?P<ids>([^,\s](?:[,\s]+)?)+))?"

# Inline 'noqa' implementation modified from flake8:
# https://github.com/PyCQA/flake8/blob/master/src/flake8/defaults.py
# We're looking for items that look like this:
# ' nosem'
# ' nosemgrep: example-pattern-id'
# ' nosem: pattern-id1,pattern-id2'
# ' NOSEMGREP:pattern-id1,pattern-id2'
#
# * We do not want to capture the ': ' that follows 'nosem'
# * We do not care about the casing of 'nosem'
# * We want a comma-separated list of ids
# * We want multi-language support, so we cannot strictly look for
#   Python comments that begin with '# '
# * nosem and nosemgrep should be interchangeable
#
NOSEM_INLINE_RE = re.compile(
    r" nosem(?:grep)?" + RULE_ID_RE_STR,
    re.IGNORECASE,
)

# A nosemgrep comment alone on its line.
# Since we don't know the comment syntax for the particular language, we
# assume it's enough that there isn't any word or number character before
# 'nosemgrep'.
# The following will not match:
#   hello(); // nosemgrep
#   + 42 // nosemgrep
# The following will match:
#   # nosemgrep
#   print('nosemgrep');
NOSEM_PREVIOUS_LINE_RE = re.compile(
    r"^[^a-zA-Z0-9]* nosem(?:grep)?" + RULE_ID_RE_STR,
    re.IGNORECASE,
)

COMMA_SEPARATED_LIST_RE = re.compile(r"[,\s]")

MAX_LINES_FLAG_NAME = "--max-lines-per-finding"
DEFAULT_MAX_LINES_PER_FINDING = 10
BREAK_LINE_WIDTH = 80
BREAK_LINE_CHAR = "-"
BREAK_LINE = BREAK_LINE_CHAR * BREAK_LINE_WIDTH

MAX_CHARS_FLAG_NAME = "--max-chars-per-line"
DEFAULT_MAX_CHARS_PER_LINE = 160
ELLIPSIS_STRING = " ... "
DEFAULT_MAX_TARGET_SIZE = 1000000  # 1 MB


class Colors(Enum):
    # these colors come from user's terminal theme
    foreground = 0
    white = 7
    black = 256
    cyan = "cyan"  # for filenames
    green = "green"  # for autofix
    yellow = "yellow"  # TODO: benchmark timing output?
    red = "red"  # for errors
    bright_blue = "bright_blue"  # TODO: line numbers?

    # these colors ignore user's terminal theme
    forced_black = 16  # #000
    forced_white = 231  # #FFF



import logging
from typing import Any
from typing import cast
from typing import Optional


class VerboseLogging(logging.Logger):
    """
    Extend logging to add a verbose logging level that is between
    INFO and DEBUG.

    Also expose a logging.verbose() method so there is no need
    to call log(VERBOSE_LEVEL, msg) every time
    """

    VERBOSE_LOG_LEVEL = 15

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def verbose(self, msg: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(self.VERBOSE_LOG_LEVEL):
            self._log(self.VERBOSE_LOG_LEVEL, msg, args, **kwargs)


def install_verbose_logging() -> None:
    """
    Makes 3 changes to stdlib logging:
    - add in logging.VERBOSE constant
    - add VERBOSE as a logging level
    - set VerboseLogging as default class returned by logging.getLogger
        - thus exposing logger.verbose(msg) method

    Any calls to getLogger before this method returns will return base
    logging.Logger class that doesn't have verbose() convenience method
    """
    # Between INFO(20) and DEBUG(10)
    logging.VERBOSE = 15  # type: ignore[attr-defined]
    logging.addLevelName(
        VerboseLogging.VERBOSE_LOG_LEVEL, "VERBOSE"
    )  # Register VERBOSE as a logging level
    logging.setLoggerClass(VerboseLogging)


install_verbose_logging()


def getLogger(name: Optional[str]) -> VerboseLogging:
    """
    Wrapper around logging.getLogger to correctly cast so mypy
    detects verbose() function
    """
    return cast(VerboseLogging, logging.getLogger(name))



#
# TODO: explain what these tests are for if you happen to know.
#
import json
import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Optional

import appdirs
import pytest

from ..conftest import chdir
from ..conftest import TESTS_PATH
from ..public_repos import ALL_LANGUAGES
from ..public_repos import ALL_REPOS

# Some improbable string that was implanted in test targets [how?] [why?].
#
SENTINEL_VALUE = 87518275812375164

LANGUAGE_SENTINELS = {
    "python": {
        "filename": "sentinel.py",
        "file_contents": f"sentinel = {SENTINEL_VALUE}",
    },
    "go": {
        "filename": "sentinel.go",
        "file_contents": f"package Sentinel\nconst sentinel = {SENTINEL_VALUE}",
    },
    "javascript": {
        "filename": "sentinel.js",
        "file_contents": f"sentinel = {SENTINEL_VALUE}",
    },
    "ruby": {
        "filename": "sentinel.rb",
        "file_contents": f"sentinel = {SENTINEL_VALUE}",
    },
}
SENTINEL_PATTERN = f"$SENTINEL = {SENTINEL_VALUE}"


def _assert_sentinel_results(
    repo_url, repo_path, sentinel_path, language, excludes=None
):
    """
    EXCLUDES: Optional[List[str]] : Each element is passed to semgrep cli as --exclude flag
    """
    cmd = [
        sys.executable,
        "-m",
        "semgrep",
        "--disable-version-check",
        "--pattern",
        SENTINEL_PATTERN,
        "--lang",
        language,
        "--json",
        repo_path,
        "--metrics",
        "off",
        "--optimizations",
        "none",  # Turn off optimizations since it skips parsing when it can and this test is testing parsing
    ]

    if excludes:
        for exclude in excludes:
            cmd.extend(["--exclude", exclude])

    semgrep_run = subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
    )

    assert semgrep_run.returncode == 0
    try:
        output = json.loads(semgrep_run.stdout)
    except json.JSONDecodeError:
        pytest.fail(
            f"Failed to parse JSON from semgrep output ({semgrep_run.args}):\n"
            + semgrep_run.stdout
            + semgrep_run.stderr
        )

    if output["errors"]:
        pytest.fail(
            f"Running on {repo_url} (cached in {repo_path}) with lang {language} had errors ({semgrep_run.args}): "
            + json.dumps(output["errors"], indent=4)
        )

    if len(output["results"]) != 1 or output["results"][0]["path"] != str(
        sentinel_path
    ):
        pytest.fail(
            f"Running on {repo_url} with lang {language} expected to have one results instead found result ({semgrep_run.args}): "
            + json.dumps(output["results"], indent=4)
        )


REPO_CACHE = Path(
    os.path.expanduser(
        os.environ.get("GITHUB_REPO_CACHE", appdirs.user_cache_dir("semgrep-tests"))
    )
)


def clone_github_repo(repo_url: str, sha: Optional[str] = None, retries: int = 3):
    """
    Internal fixture function. Do not use directly, use the `clone_github_repo` fixture.
    Wraps `_github_repo` function with retries. If the `_github_repo` throws an exception,
    it will delete `repo_destination` and retry up to `retries` times.
    """
    sha_str = sha or "latest"
    repo_dir = "-".join(repo_url.split("/")[-2:]) + "-" + sha_str
    repo_destination = REPO_CACHE / repo_dir
    try:
        return _github_repo(repo_url, sha, repo_destination)
    except (GitError, subprocess.CalledProcessError) as ex:
        print(f"Failed to clone github repo for tests {ex}")
        if repo_destination.exists():
            shutil.rmtree(repo_destination)
        if retries == 0:
            raise
        else:
            return clone_github_repo(repo_url, sha, retries - 1)


class GitError(BaseException):
    pass


def _github_repo(repo_url: str, sha: Optional[str], repo_destination: Path):
    """
    Internal fixture function. Use the `clone_github_repo` fixture.
    Clones the github repo at repo_url into `repo_destination` and checks out `sha`.

    If `repo_destination` already exists, it will validate that the correct repo is present at that location.
    """
    if not repo_destination.exists():
        if sha is None:
            subprocess.check_output(
                ["git", "clone", "--depth=1", repo_url, repo_destination]
            )
        else:
            repo_destination.mkdir()
            # Sadly, no fast way to clone a specific commit without a super
            # modern git client
            subprocess.check_output(["git", "clone", repo_url, repo_destination])
            with chdir(repo_destination):
                subprocess.check_output(["git", "checkout", sha])

    # validate that the repo seems setup properly
    with chdir(repo_destination):
        # some tests modify it, lets put everything back to normal
        subprocess.check_output(["git", "clean", "-fd"])
        subprocess.check_output(["git", "reset", "--hard"])
        all_clean = (
            subprocess.check_output(["git", "status", "--porcelain"]).strip() == b""
        )
        if not all_clean:
            raise GitError("Couldn't clean the repo, something is wrong. Deleting.")
        repo_sha = subprocess.check_output(["git", "rev-parse", "HEAD"])
        if sha:
            if not repo_sha.startswith(sha.encode("utf-8")):
                shutil.rmtree(repo_destination)
                raise GitError(
                    f"Github repo is broken (not set to correct sha: {repo_sha.decode('utf-8', errors='replace')}"
                )

    return repo_destination


@pytest.mark.parametrize("repo_object", ALL_REPOS)
#
# This test runs [which checks?] against one public git repo.
# See the list of repos in public_repos.py.
# Those repos, once downloaded are cached locally in ~/.cache.
# You may have to clear the cache manually if some tests start failing.
#
# A repo_object is a GitHub URL, a list of applicable languages, and whether
# the test is supposed to pass or fail. The expectation ok/xfail is
# consulted outside of this function. [where?]
#
def test_semgrep_on_repo(monkeypatch, tmp_path, repo_object):
    (tmp_path / "rules").symlink_to(Path(TESTS_PATH / "qa" / "rules").resolve())

    monkeypatch.chdir(tmp_path)

    repo_url = repo_object["repo"]
    languages = repo_object["languages"]
    excludes = repo_object.get("excludes")
    repo_path = clone_github_repo(repo_url=repo_url)
    repo_languages = (
        LANGUAGE_SENTINELS
        if languages == ALL_LANGUAGES
        else {
            language: sentinel_info
            for language, sentinel_info in LANGUAGE_SENTINELS.items()
            if language in languages
        }
    )

    for language, sentinel_info in repo_languages.items():
        sentinel_path = repo_path / sentinel_info["filename"]
        with sentinel_path.open("w") as sentinel_file:
            sentinel_file.write(sentinel_info["file_contents"])

        _assert_sentinel_results(repo_url, repo_path, sentinel_path, language, excludes)

    cmd = [
        sys.executable,
        "-m",
        "semgrep",
        "--disable-version-check",
        "--config=rules/regex-sentinel.yaml",
        "--strict",
        "--json",
        "--metrics",
        "off",
        "--optimizations",
        "none",  # Turn off optimizations since it skips parsing when it can and this test is testing parsing
        repo_path,
    ]

    if excludes:
        for exclude in excludes:
            cmd.extend(["--exclude", exclude])

    print(f"semgrep command: {cmd}")

    res = subprocess.run(
        cmd,
        encoding="utf-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    returncode = res.returncode
    print("--- semgrep error output ---")
    print(res.stderr)
    print("----------------------------")
    print("--- semgrep standard output ---")
    print(res.stdout)
    print("-------------------------------")
    if returncode != 0:
        # Fail regardless of the expected status "ok" or "xfail".
        raise subprocess.SubprocessError(
            f"Semgrep exited with non-zero status: {returncode}"
        )
    output = json.loads(res.stdout)

    expected_results_count = len(repo_languages)
    if len(output["results"]) != expected_results_count or len(output["errors"]) != 0:
        pytest.fail(
            f"Running on {repo_url} with regex rules. Expect {expected_results_count} results and no errors but got: "
            + json.dumps(output, indent=4)
        )



import json
import os
import time
from json import JSONDecodeError
from pathlib import Path
from typing import Mapping
from typing import Optional

from packaging.version import InvalidVersion
from packaging.version import Version

from semgrep import __VERSION__
from semgrep.constants import SEMGREP_USER_AGENT
from semgrep.types import JsonObject
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)

VERSION_CHECK_URL = str(
    os.environ.get("SEMGREP_VERSION_CHECK_URL", "https://semgrep.dev/api/check-version")
)
VERSION_CHECK_TIMEOUT = int(
    os.environ.get(
        "SEMGREP_VERSION_CHECK_TIMEOUT", 2  # Don't block user's for too long
    )
)
VERSION_CACHE_PATH = Path(
    os.environ.get(
        "SEMGREP_VERSION_CACHE_PATH",
        Path.home() / ".cache" / "semgrep_version",
    )
)


def _fetch_latest_version(
    url: str = VERSION_CHECK_URL, timeout: int = VERSION_CHECK_TIMEOUT
) -> Optional[JsonObject]:
    try:
        import requests

        resp = requests.get(
            url,
            headers={"User-Agent": SEMGREP_USER_AGENT},
            timeout=timeout,
        )
    except Exception as e:
        logger.debug(f"Fetching latest version failed to connect: {e}")
        return None

    if resp.status_code != requests.codes.OK:
        logger.debug(
            f"Fetching latest version received HTTP error code: {resp.status_code}"
        )
        return None
    try:
        res = resp.json()
    except ValueError:
        logger.debug("Fetching latest version received invalid JSON")
        return None

    if not isinstance(res, Mapping):
        logger.debug("Latest version response is not an object")
        return None

    return res


def _get_version_from_cache(version_cache_path: Path) -> Optional[JsonObject]:
    now = time.time()

    if not version_cache_path.is_file():
        logger.debug("Version cache does not exist")
        return None

    with version_cache_path.open() as f:
        timestamp_str = f.readline().strip()
        latest_version_str = f.readline().strip()

    try:
        # Treat time as integer seconds so no need to deal with str float conversion
        timestamp = int(timestamp_str)
    except ValueError:
        logger.debug(f"Version cache invalid timestamp: {timestamp_str}")
        return None

    one_day = 86400
    if now - timestamp > one_day:
        logger.debug(f"Version cache expired: {timestamp_str}:{now}")
        return None

    try:
        res = json.loads(latest_version_str)
    except JSONDecodeError:
        logger.debug(f"Version cache does not contain JSON object")
        return None

    if not isinstance(res, Mapping):
        logger.debug("Latest version response is not an object")
        return None

    return res


def _get_latest_version(version_cache_path: Path) -> Optional[JsonObject]:
    latest_version = _get_version_from_cache(version_cache_path)

    if latest_version is None:
        latest_version = _fetch_latest_version()

    if latest_version is None:
        # Request timed out or invalid
        return None

    version_cache_path.parent.mkdir(parents=True, exist_ok=True)
    with version_cache_path.open("w") as f:
        # Integer time so no need to deal with str float conversions
        f.write(f"{int(time.time())}\n")
        f.write(json.dumps(latest_version))

    return latest_version


def _show_banners(current_version: Version, latest_version_object: JsonObject) -> None:
    banners = latest_version_object.get("banners", [])
    for b in banners:
        try:
            show_str = b.get("show_version")  # Note that b["show_version"] can be None
            show = Version(show_str) if show_str else None
            hide_str = b.get("hide_version")
            hide = Version(hide_str) if hide_str else None
        except InvalidVersion as e:
            logger.debug(f"Invalid version string: {e}")
            continue
        if (not show or current_version >= show) and (
            not hide or current_version < hide
        ):
            logger.warning("\n" + b.get("message", ""))


def version_check(version_cache_path: Path = VERSION_CACHE_PATH) -> None:
    """
    Checks for messages from the backend, displaying any messages that match the current version

    :param version_cache_path: Path where we cache the backend response
    """
    latest_version_object = _get_latest_version(version_cache_path)
    if latest_version_object is None:
        return

    try:
        current_version = Version(__VERSION__)
    except InvalidVersion as e:
        logger.debug(f"Invalid version string: {e}")
        return

    _show_banners(current_version, latest_version_object)



from typing import Any
from typing import Iterable
from typing import Mapping
from typing import Sequence

from semgrep.constants import CLI_RULE_ID
from semgrep.error import SemgrepError
from semgrep.formatter.base import BaseFormatter
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch


class EmacsFormatter(BaseFormatter):
    @staticmethod
    def _get_parts(rule_match: RuleMatch) -> Sequence[str]:
        check_id = (
            rule_match.id.split(".")[-1] if rule_match.id != CLI_RULE_ID else None
        )
        match_severity = rule_match.severity.value.lower()
        severity = match_severity + f"({check_id})" if check_id else match_severity
        return [
            str(rule_match.path),
            str(rule_match.start.line),
            str(rule_match.start.col),
            severity,
            rule_match.lines[0].rstrip(),
            rule_match.message,
        ]

    def format(
        self,
        rules: Iterable[Rule],
        rule_matches: Iterable[RuleMatch],
        semgrep_structured_errors: Sequence[SemgrepError],
        extra: Mapping[str, Any],
    ) -> str:
        sorted_matches = sorted(rule_matches, key=lambda r: (r.path, r.id))
        return "\n".join(":".join(self._get_parts(rm)) for rm in sorted_matches)



#!/usr/bin/env python3
import click

from semgrep.commands.login import login
from semgrep.commands.login import logout
from semgrep.commands.scan import scan
from semgrep.default_group import DefaultGroup


@click.group(cls=DefaultGroup, default_command="scan")
def cli() -> None:
    pass


cli.add_command(login)
cli.add_command(logout)
cli.add_command(scan)



import hashlib
import json
from copy import deepcopy
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Union

from semgrep.constants import RuleSeverity
from semgrep.error import InvalidRuleSchemaError
from semgrep.rule_lang import EmptySpan
from semgrep.rule_lang import RuleValidation
from semgrep.rule_lang import Span
from semgrep.rule_lang import YamlMap
from semgrep.rule_lang import YamlTree
from semgrep.semgrep_types import ALLOWED_GLOB_TYPES
from semgrep.semgrep_types import JOIN_MODE
from semgrep.semgrep_types import LANGUAGE
from semgrep.semgrep_types import Language
from semgrep.semgrep_types import SEARCH_MODE
from semgrep.util import flatten


class Rule:
    def __init__(self, raw: YamlTree[YamlMap]) -> None:
        self._yaml = raw
        self._raw: Dict[str, Any] = raw.unroll_dict()

        self._id = str(self._raw["id"])

        paths_tree: Optional[YamlTree] = self._yaml.value.get("paths")
        if paths_tree is None:
            path_dict = {}
        else:
            paths, paths_span = paths_tree.value, paths_tree.span
            if not isinstance(paths, YamlMap):
                path_key = self._yaml.value.key_tree("paths").span
                help_str: Optional[str] = None
                if isinstance(paths, list):
                    help_str = "remove the `-` to convert the list into a mapping"
                raise InvalidRuleSchemaError(
                    short_msg="invalid paths",
                    long_msg=f"the `paths:` targeting rules must be an object with at least one of {ALLOWED_GLOB_TYPES}",
                    spans=[path_key.extend_to(paths_span)],
                    help=help_str,
                )
            path_dict = paths_tree.unroll_dict()
        self._includes = cast(Sequence[str], path_dict.get("include", []))
        self._excludes = cast(Sequence[str], path_dict.get("exclude", []))
        rule_languages: Set[Language] = {
            LANGUAGE.resolve(l, self.languages_span)
            for l in self._raw.get("languages", [])
        }

        # add typescript to languages if the rule supports javascript.
        # TODO: Move this hack to lang.json
        if any(
            language == LANGUAGE.resolve("javascript") for language in rule_languages
        ):
            rule_languages.add(LANGUAGE.resolve("typescript"))
            self._raw["languages"] = [str(l) for l in rule_languages]

        self._languages = sorted(rule_languages)

        # check taint/search mode
        if self._raw.get("mode") == JOIN_MODE:
            self._mode = JOIN_MODE
        else:
            self._mode = SEARCH_MODE

        # TODO: Move this hack to lang.json
        if any(language == LANGUAGE.resolve("regex") for language in self._languages):
            self._validate_none_language_rule()

    def _validate_none_language_rule(self) -> None:
        """
        For regex-only rules, only patterns, pattern-either, and pattern-regex is valid.
        """

        def _recursive_contains(
            obj: Union[Dict[str, Any], List[Any], str], search_key: str
        ) -> bool:
            """
            Returns true if object contains any object that contains search_key as key
            """
            if isinstance(obj, dict):
                for key in obj:
                    if key == search_key:
                        return True

                    if _recursive_contains(obj[key], search_key):
                        return True

            if isinstance(obj, list):
                for elem in obj:
                    if _recursive_contains(elem, search_key):
                        return True

            return False

        if _recursive_contains(self._raw, "pattern"):
            raise InvalidRuleSchemaError(
                short_msg=f"invalid pattern clause",
                long_msg=f"invalid pattern clause 'pattern' with regex-only rules in rule: {self.id}",
                spans=[],
                help=f"use only patterns, pattern-either, pattern-regex, or pattern-not-regex with regex-only rules",
            )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False

        return self._raw == other._raw

    def __hash__(self) -> int:
        return hash(self.id)

    @property
    def includes(self) -> Sequence[str]:
        return self._includes

    @property
    def excludes(self) -> Sequence[str]:
        return self._excludes

    @property
    def id(self) -> str:
        return self._id

    @property
    def message(self) -> str:
        return str(self._raw["message"])

    @property
    def metadata(self) -> Dict[str, Any]:
        return self._raw.get("metadata", {})

    @property
    def severity(self) -> RuleSeverity:
        return RuleSeverity(self._raw["severity"])

    @property
    def mode(self) -> str:
        return self._mode

    @property
    def project_depends_on(self) -> Optional[List[List[Dict[str, str]]]]:
        """
        If the rule contains `project-depends-on` keys under patterns, return the values of those keys
        Otherwise return None
        """
        # TODO: in initial implementation, this key is allowed only as a top-level key under `patterns`
        PROJECT_DEPENDS_ON_KEY_NAME = "r2c-internal-project-depends-on"
        matched_keys = [d for d in self._raw.get("patterns", [])]
        depends_entries = [
            list(_.values()) for _ in matched_keys if PROJECT_DEPENDS_ON_KEY_NAME in _
        ]
        flattened = flatten(depends_entries)
        if len(flattened) == 0:
            return None
        return flattened

    @property
    def languages(self) -> List[Language]:
        return self._languages

    @property
    def languages_span(self) -> Span:
        return self._yaml.value["languages"].span

    @property
    def raw(self) -> Dict[str, Any]:
        return self._raw

    @property
    def fix(self) -> Optional[str]:
        return self._raw.get("fix")

    @property
    def fix_regex(self) -> Optional[Dict[str, Any]]:
        return self._raw.get("fix-regex")

    @classmethod
    def from_json(cls, rule_json: Dict[str, Any]) -> "Rule":
        yaml = YamlTree.wrap(rule_json, EmptySpan)
        return cls(yaml)

    @classmethod
    def from_yamltree(cls, rule_yaml: YamlTree[YamlMap]) -> "Rule":
        return cls(rule_yaml)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id}>"

    def rename_id(self, new_id: str) -> None:
        self._id = new_id
        self._raw["id"] = new_id

    @property
    def full_hash(self) -> str:
        """
        sha256 hash of the whole rule object instead of just the id
        """
        return hashlib.sha256(
            json.dumps(self._raw, sort_keys=True).encode()
        ).hexdigest()

    @property
    def should_run_on_semgrep_core(self) -> bool:
        """
        Used to detect whether the rule had patterns that need to run on the core
        (beyond Python-handled patterns, like `pattern-depends-on`).
        Remove this code once all rule runnning is done in the core and the answer is always 'yes'
        """

        def has_runnable_rule(d: Dict[str, Any]) -> bool:
            for k in d:
                if k in RuleValidation.PATTERN_KEYS:
                    children = d.get(k)
                    if children is not None and isinstance(children, list):
                        return any(has_runnable_rule(_) for _ in children)
                    return True
            return False

        return has_runnable_rule(self._raw)


def rule_without_metadata(rule: Rule) -> Rule:
    rule2 = deepcopy(rule)
    rule2._raw["metadata"] = {}
    return rule2



import re
import subprocess


def test_version():
    result = subprocess.check_output(
        ["semgrep", "--version", "--disable-version-check"], encoding="utf-8"
    )

    assert re.match(r"\d+\.\d+\.\d+", result)



import json
from typing import Any
from typing import Iterable
from typing import Mapping
from typing import Sequence

from semgrep.error import SemgrepError
from semgrep.formatter.base import BaseFormatter
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch


class JsonFormatter(BaseFormatter):
    @staticmethod
    def _rule_match_to_json(rule_match: RuleMatch) -> Mapping[str, Any]:
        extra = rule_match.extra
        extra["message"] = rule_match.message
        extra["metadata"] = rule_match.metadata
        extra["severity"] = rule_match.severity.value

        # 'lines' already contains '\n' at the end of each line
        extra["lines"] = "".join(rule_match.lines).rstrip()

        if rule_match.fix:
            extra["fix"] = rule_match.fix
        if rule_match.fix_regex:
            extra["fix_regex"] = rule_match.fix_regex
        if rule_match.is_ignored is not None:
            extra["is_ignored"] = rule_match.is_ignored

        return {
            "check_id": rule_match.id,
            "path": str(rule_match.path),
            "start": rule_match.start.to_dict(),
            "end": rule_match.end.to_dict(),
            "extra": extra,
        }

    def format(
        self,
        rules: Iterable[Rule],
        rule_matches: Iterable[RuleMatch],
        semgrep_structured_errors: Sequence[SemgrepError],
        extra: Mapping[str, Any],
    ) -> str:
        output_dict = {
            "results": [
                self._rule_match_to_json(rule_match) for rule_match in rule_matches
            ],
            "errors": [error.to_dict() for error in semgrep_structured_errors],
            **extra,
        }
        # Sort keys for predictable output. This helps with snapshot tests, etc.
        return json.dumps(output_dict, sort_keys=True)
